"""
==================================
Map two radars to a Cartesian grid
==================================

Map the reflectivity field of two nearby ARM XSARP radars from antenna
coordinates to a Cartesian grid.

"""
print(__doc__)

# Author: Jonathan J. Helmus (jhelmus@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt

import pyart
from pyart.testing import get_test_data

# read in the data from both XSAPR radars
xsapr_sw_file = get_test_data("swx_20120520_0641.nc")
xsapr_se_file = get_test_data("sex_20120520_0641.nc")
radar_sw = pyart.io.read_cfradial(xsapr_sw_file)
radar_se = pyart.io.read_cfradial(xsapr_se_file)

# filter out gates with reflectivity > 100 from both radars
gatefilter_se = pyart.filters.GateFilter(radar_se)
gatefilter_se.exclude_transition()
gatefilter_se.exclude_above("corrected_reflectivity_horizontal", 100)
gatefilter_sw = pyart.filters.GateFilter(radar_sw)
gatefilter_sw.exclude_transition()
gatefilter_sw.exclude_above("corrected_reflectivity_horizontal", 100)

# perform Cartesian mapping, limit to the reflectivity field.
grid = pyart.map.grid_from_radars(
    (radar_se, radar_sw),
    gatefilters=(gatefilter_se, gatefilter_sw),
    grid_shape=(1, 201, 201),
    grid_limits=((1000, 1000), (-50000, 40000), (-60000, 40000)),
    grid_origin=(36.57861, -97.363611),
    fields=["corrected_reflectivity_horizontal"],
)

# create the plot
fig = plt.figure()
ax = fig.add_subplot(111)
ax.imshow(
    grid.fields["corrected_reflectivity_horizontal"]["data"][0],
    origin="lower",
    extent=(-60, 40, -50, 40),
    vmin=0,
    vmax=48,
)
plt.show()



"""
=======================================
Create an image-muted reflectivity plot
=======================================
An example which creates an image-muted PPI plot from a NEXRAD file.

Image muting reduces the visual prominence of the reflectivities within identified
melting and mixed precipitation features in winter storms (i.e. regions with low
correlation coefficient values). Reflectivities corresponding to melting and mixed
precipitation features are deemphasized using a gray scale and the regions
with just snow and just rain are depicted in a corresponding full-color scale.
The ultimate utility of image muting radar reflectivity is to reduce the misinterpretation
of regions of melting or mixed precipitation as opposed to heavy snow or heavy rain.

"""

print(__doc__)

# Author: Laura Tomkins (lmtomkin@ncsu.edu)
# License: BSD 3 clause
# citation: Tomkins, L. M., Yuter, S. E., Miller, M. A., and Allen, L. R., 2022:
# Image muting of mixed precipitation to improve identification of regions
# of heavy snow in radar data. Atmos. Meas. Tech., 15, 5515–5525,
# https://doi.org/10.5194/amt-15-5515-2022

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np

import pyart

# Read in file
nexrad_file = "s3://noaa-nexrad-level2/2020/02/07/KBGM/KBGM20200207_132642_V06"
radar = pyart.io.read_nexrad_archive(nexrad_file)

# Mute radar object
# Regions where rhoHV < 0.97 and reflectivity > 20 will be muted
radar = pyart.util.image_mute_radar(
    radar,
    field="reflectivity",
    mute_field="cross_correlation_ratio",
    mute_threshold=0.97,
    field_threshold=20,
)

# adjust colormaps for visual separation
# this example uses perceptually uniform colormaps
magma_cmap = plt.get_cmap("magma_r")
grays_cmap = plt.get_cmap("gray_r")

nonmuted_cmap = mcolors.LinearSegmentedColormap.from_list(
    "nonmuted_cmap", magma_cmap(np.linspace(0, 0.9, magma_cmap.N))
)
muted_cmap = mcolors.LinearSegmentedColormap.from_list(
    "muted_cmap", grays_cmap(np.linspace(0, 0.7, grays_cmap.N))
)

# create plot using RadarDisplay
display = pyart.graph.RadarDisplay(radar)

fig = plt.figure()
ax = plt.axes()
display.plot("nonmuted_reflectivity", 0, vmin=5, vmax=45, cmap=nonmuted_cmap)
display.plot("muted_reflectivity", 0, vmin=5, vmax=45, cmap=muted_cmap)
display.set_limits((-300, 300), (-300, 300))
ax.set_aspect("equal")
plt.show()



""" Unit tests for rainfall rate estimation module. """

from numpy.testing import assert_allclose

import pyart


def test_est_rain_rate_zpoly():
    radar = pyart.io.read(pyart.testing.UF_FILE)
    rain = pyart.retrieve.est_rain_rate_zpoly(radar)

    assert "units" in rain.keys()
    assert "long_name" in rain.keys()
    assert "coordinates" in rain.keys()
    assert rain["units"] == "mm/hr"

    assert_allclose(
        rain["data"][0][0:5],
        [
            2.89949111e-04,
            1.26031121e-02,
            9.54675995e-06,
            2.14264704e-01,
            9.47527914e-01,
        ],
        atol=1e7,
    )

    assert_allclose(
        rain["data"][0][-5:],
        [0.20666919, 0.10129632, 0.13916402, 0.12526962, 0.12681702],
        atol=1e7,
    )
    del radar


def test_est_rain_rate_z():
    radar = pyart.io.read(pyart.testing.UF_FILE)
    rain = pyart.retrieve.est_rain_rate_z(radar)

    assert "units" in rain.keys()
    assert "long_name" in rain.keys()
    assert "coordinates" in rain.keys()
    assert rain["units"] == "mm/hr"

    assert_allclose(
        rain["data"][0][0:5],
        [0.01604766, 0.05375671, 0.00767614, 0.27197879, 1.04875935],
        atol=1e7,
    )

    assert_allclose(
        rain["data"][0][-5:],
        [0.26443019, 0.15909914, 0.1973247, 0.18339988, 0.18495507],
        atol=1e7,
    )
    del radar


def test_est_rain_rate_kdp():
    radar = pyart.io.read(pyart.testing.UF_FILE)
    rain = pyart.retrieve.est_rain_rate_kdp(radar)

    assert "units" in rain.keys()
    assert "long_name" in rain.keys()
    assert "coordinates" in rain.keys()
    assert rain["units"] == "mm/hr"

    assert_allclose(rain["data"][0][0:5], [0.0, 0.0, 0.0, 0.0, 0.0], atol=1e7)

    assert_allclose(
        rain["data"][0][-5:],
        [57.4029591, 56.8539739, 56.57898183, 56.30365354, 56.11991333],
        atol=1e7,
    )

    radar.instrument_parameters.pop("frequency")
    rain_no_freq = pyart.retrieve.est_rain_rate_kdp(radar)

    assert_allclose(rain_no_freq["data"][0][0:5], [0.0, 0.0, 0.0, 0.0, 0.0], atol=1e7)

    assert_allclose(
        rain_no_freq["data"][0][-5:],
        [117.04546973, 115.85529152, 115.25939369, 114.66295169, 114.26501948],
        atol=1e7,
    )
    del radar


def test_est_rain_rate_a():
    radar = pyart.io.read(pyart.testing.UF_FILE)
    spec_a, _ = pyart.correct.calculate_attenuation(radar, z_offset=1.0)
    radar.add_field("specific_attenuation", spec_a, replace_existing=True)
    rain = pyart.retrieve.est_rain_rate_a(radar)

    assert "units" in rain.keys()
    assert "long_name" in rain.keys()
    assert "coordinates" in rain.keys()
    assert rain["units"] == "mm/hr"

    assert_allclose(
        rain["data"][0][0:5],
        [0.03885328, 0.02587043, 0.02171088, 0.13686746, 1.2150548],
        atol=1e7,
    )

    assert_allclose(rain["data"][0][-5:], [0.0, 0.0, 0.0, 0.0, 0.0], atol=1e7)

    radar.instrument_parameters.pop("frequency")
    rain_no_freq = pyart.retrieve.est_rain_rate_kdp(radar)

    assert_allclose(
        rain_no_freq["data"][0][0:5],
        [0.10804109, 0.06917369, 0.05707911, 0.42970675, 4.708362],
        atol=1e7,
    )

    assert_allclose(rain_no_freq["data"][0][-5:], [0.0, 0.0, 0.0, 0.0, 0.0], atol=1e7)
    del radar


def test_coeff_ra_table():
    coeff_ra_dict = pyart.retrieve.qpe._coeff_ra_table()
    test_dict = {"S": (3100.0, 1.03), "C": (250.0, 0.91), "X": (45.5, 0.83)}

    assert coeff_ra_dict == test_dict


def test_coeff_rkdp_table():
    coeff_rkdp_dict = pyart.retrieve.qpe._coeff_rkdp_table()
    test_dict = {"S": (50.7, 0.85), "C": (29.7, 0.85), "X": (15.81, 0.7992)}

    assert coeff_rkdp_dict == test_dict


def test_get_coeff_ra():
    coeff_ra_good = pyart.retrieve.qpe._get_coeff_ra(3e9)
    assert coeff_ra_good == (3100.0, 1.03)

    coeff_ra_use_s = pyart.retrieve.qpe._get_coeff_ra(1e9)
    assert coeff_ra_use_s == (3100.0, 1.03)

    coeff_ra_use_x = pyart.retrieve.qpe._get_coeff_ra(13e9)
    assert coeff_ra_use_x == (45.5, 0.83)


def test_get_coeff_rkdp():
    coeff_rkdp_good = pyart.retrieve.qpe._get_coeff_rkdp(3e9)
    assert coeff_rkdp_good == (50.7, 0.85)

    coeff_rkdp_use_s = pyart.retrieve.qpe._get_coeff_rkdp(1e9)
    assert coeff_rkdp_use_s == (50.7, 0.85)

    coeff_rkdp_use_x = pyart.retrieve.qpe._get_coeff_rkdp(13e9)
    assert coeff_rkdp_use_x == (15.81, 0.7992)



"""
Functions for creating sample Radar and Grid objects.

"""

import numpy as np
import scipy

from ..config import get_metadata
from ..core.grid import Grid
from ..core.radar import Radar
from ..exceptions import MissingOptionalDependency
from .sample_files import _EXAMPLE_RAYS_FILE

try:
    import xarray as xr

    from ..core.radar_spectra import RadarSpectra

    _XARRAY_AVAILABLE = True
except ImportError:
    _XARRAY_AVAILABLE = False


def make_empty_ppi_radar(ngates, rays_per_sweep, nsweeps):
    """
    Return an Radar object, representing a PPI scan.

    Parameters
    ----------
    ngates : int
        Number of gates per ray.
    rays_per_sweep : int
        Number of rays in each PPI sweep.
    nsweeps : int
        Number of sweeps.

    Returns
    -------
    radar : Radar
        Radar object with no fields, other parameters are set to default
        values.

    """
    nrays = rays_per_sweep * nsweeps

    time = get_metadata("time")
    _range = get_metadata("range")
    latitude = get_metadata("latitude")
    longitude = get_metadata("longitude")
    altitude = get_metadata("altitude")
    sweep_number = get_metadata("sweep_number")
    sweep_mode = get_metadata("sweep_mode")
    fixed_angle = get_metadata("fixed_angle")
    sweep_start_ray_index = get_metadata("sweep_start_ray_index")
    sweep_end_ray_index = get_metadata("sweep_end_ray_index")
    azimuth = get_metadata("azimuth")
    elevation = get_metadata("elevation")

    fields = {}
    scan_type = "ppi"
    metadata = {"instrument_name": "fake_radar"}

    time["data"] = np.arange(nrays, dtype="float64")
    time["units"] = "seconds since 1989-01-01T00:00:01Z"
    _range["data"] = np.linspace(0, 1000, ngates).astype("float32")

    latitude["data"] = np.array([36.5], dtype="float64")
    longitude["data"] = np.array([-97.5], dtype="float64")
    altitude["data"] = np.array([200], dtype="float64")

    sweep_number["data"] = np.arange(nsweeps, dtype="int32")
    sweep_mode["data"] = np.array(["azimuth_surveillance"] * nsweeps)
    fixed_angle["data"] = np.array([0.75] * nsweeps, dtype="float32")
    sweep_start_ray_index["data"] = np.arange(0, nrays, rays_per_sweep, dtype="int32")
    sweep_end_ray_index["data"] = np.arange(
        rays_per_sweep - 1, nrays, rays_per_sweep, dtype="int32"
    )

    azimuth["data"] = np.arange(nrays, dtype="float32")
    elevation["data"] = np.array([0.75] * nrays, dtype="float32")

    return Radar(
        time,
        _range,
        fields,
        metadata,
        scan_type,
        latitude,
        longitude,
        altitude,
        sweep_number,
        sweep_mode,
        fixed_angle,
        sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth,
        elevation,
        instrument_parameters=None,
    )


def make_empty_rhi_radar(ngates, rays_per_sweep, nsweeps):
    """
    Return an Radar object, representing a RHI scan.

    Parameters
    ----------
    ngates : int
        Number of gates per ray.
    rays_per_sweep : int
        Number of rays in each PPI sweep.
    nsweeps : int
        Number of sweeps.

    Returns
    -------
    radar : Radar
        Radar object with no fields, other parameters are set to default
        values.

    """
    radar = make_empty_ppi_radar(ngates, rays_per_sweep, nsweeps)
    radar.scan_type = "rhi"
    nrays = rays_per_sweep * nsweeps
    radar.sweep_mode["data"] = np.array(["rhi"] * nsweeps)
    radar.elevation["data"] = np.arange(nrays, dtype="float32")
    radar.azimuth["data"] = np.array([0.75] * nrays, dtype="float32")
    return radar


def make_target_radar():
    """
    Return a PPI radar with a target like reflectivity field.
    """
    radar = make_empty_ppi_radar(50, 360, 1)
    fields = {"reflectivity": get_metadata("reflectivity")}
    fdata = np.zeros((360, 50), dtype="float32")
    fdata[:, 0:10] = 0.0
    fdata[:, 10:20] = 10.0
    fdata[:, 20:30] = 20.0
    fdata[:, 30:40] = 30.0
    fdata[:, 40:50] = 40.0
    fields["reflectivity"]["data"] = fdata
    radar.fields = fields
    return radar


def make_velocity_aliased_radar(alias=True):
    """
    Return a PPI radar with a target like reflectivity field.

    Set alias to False to return a de-aliased radar.
    """
    radar = make_empty_ppi_radar(50, 360, 1)
    radar.range["meters_between_gates"] = 1.0
    radar.range["meters_to_center_of_first_gate"] = 1.0
    radar.instrument_parameters = {"nyquist_velocity": {"data": np.array([10.0] * 360)}}

    fields = {
        "reflectivity": get_metadata("reflectivity"),
        "velocity": get_metadata("velocity"),
    }

    # fake reflectivity data, all zero reflectivity
    fdata = np.zeros((360, 50), dtype="float32")
    fields["reflectivity"]["data"] = fdata

    # fake velocity data, all zeros except a wind burst on at ~13 degrees.
    # burst is partially aliased.
    vdata = np.zeros((360 * 1, 50), dtype="float32")

    for i, idx in enumerate(range(13, -1, -1)):
        vdata[i, idx : idx + i + 1] = np.arange(0.5, 0.5 + i * 1.0 + 0.001)
    vdata[:14, 14:27] = vdata[:14, 12::-1]  # left/right flip
    vdata[14:27] = vdata[12::-1, :]  # top/bottom flip
    aliased = np.where(vdata > 10.0)
    if alias:
        vdata[aliased] += -20.0
    fields["velocity"]["data"] = vdata

    radar.fields = fields
    return radar


def make_velocity_aliased_rhi_radar(alias=True):
    """
    Return a RHI radar with a target like reflectivity field.

    Set alias to False to return a de-aliased radar.
    """
    radar = make_empty_rhi_radar(50, 180, 1)
    radar.range["meters_between_gates"] = 1.0
    radar.range["meters_to_center_of_first_gate"] = 1.0
    radar.instrument_parameters = {"nyquist_velocity": {"data": np.array([10.0] * 360)}}

    fields = {
        "reflectivity": get_metadata("reflectivity"),
        "velocity": get_metadata("velocity"),
    }

    # fake reflectivity data, all zero reflectivity
    fdata = np.zeros((180, 50), dtype="float32")
    fields["reflectivity"]["data"] = fdata

    # fake velocity data, all zeros except a wind burst on at ~13 degrees.
    # burst is partially aliased.
    vdata = np.zeros((180 * 1, 50), dtype="float32")

    for i, idx in enumerate(range(13, -1, -1)):
        vdata[i, idx : idx + i + 1] = np.arange(0.5, 0.5 + i * 1.0 + 0.001)
    vdata[:14, 14:27] = vdata[:14, 12::-1]  # left/right flip
    vdata[14:27] = vdata[12::-1, :]  # top/bottom flip
    aliased = np.where(vdata > 10.0)
    if alias:
        vdata[aliased] += -20.0
    fields["velocity"]["data"] = vdata

    radar.fields = fields
    return radar


def make_single_ray_radar():
    """
    Return a PPI radar with a single ray taken from a ARM C-SAPR Radar.

    Radar object returned has 'reflectivity_horizontal',
    'norm_coherent_power', 'copol_coeff', 'dp_phase_shift', 'diff_phase', and
    'differential_reflectivity' fields with no metadata but a 'data' key.
    This radar is used for unit tests in correct modules.

    """
    radar = make_empty_ppi_radar(983, 1, 1)
    radar.range["data"] = 117.8784 + np.arange(983) * 119.91698
    f = np.load(_EXAMPLE_RAYS_FILE)
    for field_name in f:
        radar.fields[field_name] = {"data": f[field_name]}
    f.close()
    return radar


def make_empty_grid(grid_shape, grid_limits):
    """
    Make an empty grid object without any fields or metadata.

    Parameters
    ----------
    grid_shape : 3-tuple of floats
        Number of points in the grid (z, y, x).
    grid_limits : 3-tuple of 2-tuples
        Minimum and maximum grid location (inclusive) in meters for the
        z, y, x coordinates.

    Returns
    -------
    grid : Grid
        Empty Grid object, centered near the ARM SGP site (Oklahoma).

    """
    time = get_metadata("grid_time")
    time["data"] = np.array([0.0])
    time["units"] = "seconds since 2000-01-01T00:00:00Z"

    # grid coordinate dictionaries
    nz, ny, nx = grid_shape
    (z0, z1), (y0, y1), (x0, x1) = grid_limits

    x = get_metadata("x")
    x["data"] = np.linspace(x0, x1, nx)

    y = get_metadata("y")
    y["data"] = np.linspace(y0, y1, ny)

    z = get_metadata("z")
    z["data"] = np.linspace(z0, z1, nz)

    origin_altitude = get_metadata("origin_altitude")
    origin_altitude["data"] = np.array([300.0])

    origin_latitude = get_metadata("origin_latitude")
    origin_latitude["data"] = np.array([36.74])

    origin_longitude = get_metadata("origin_longitude")
    origin_longitude["data"] = np.array([-98.1])

    fields = {}
    metadata = {}

    radar_latitude = get_metadata("radar_latitude")
    radar_latitude["data"] = np.array([36.74])

    radar_longitude = get_metadata("radar_longitude")
    radar_longitude["data"] = np.array([-98.1])

    radar_altitude = get_metadata("radar_altitude")
    radar_altitude["data"] = np.array([300.0])

    radar_time = get_metadata("radar_time")
    radar_time["data"] = np.array([0.0])
    radar_time["units"] = "seconds since 2000-01-01T00:00:00Z"

    radar_name = get_metadata("radar_name")
    radar_name["data"] = np.array(["ExampleRadar"])

    return Grid(
        time,
        fields,
        metadata,
        origin_latitude,
        origin_longitude,
        origin_altitude,
        x,
        y,
        z,
        radar_latitude=radar_latitude,
        radar_longitude=radar_longitude,
        radar_altitude=radar_altitude,
        radar_time=radar_time,
        radar_name=radar_name,
    )


def make_target_grid():
    """
    Make a sample Grid with a rectangular target.
    """
    grid_shape = (2, 400, 320)
    grid_limits = ((0, 500), (-400000, 400000), (-300000, 300000))
    grid = make_empty_grid(grid_shape, grid_limits)

    fdata = np.zeros((2, 400, 320), dtype="float32")
    fdata[:, 50:-50, 40:-40] = 10.0
    fdata[:, 100:-100, 80:-80] = 20.0
    fdata[:, 150:-150, 120:-120] = 30.0
    fdata[1] += 5
    rdic = {"data": fdata, "long_name": "reflectivity", "units": "dBz"}
    grid.fields = {"reflectivity": rdic}
    return grid


def make_storm_grid():
    """
    Make a sample Grid with a rectangular storm target.
    """
    grid_shape = (2, 50, 40)
    grid_limits = ((0, 500), (-400000, 400000), (-300000, 300000))
    grid = make_empty_grid(grid_shape, grid_limits)

    fdata = np.ma.zeros((2, 50, 40), dtype="float32")
    fdata[:] = np.ma.masked
    fdata[:, 5:-5, 4:-4] = 20.0
    fdata[:, 10:-10, 8:-8] = 30.0
    fdata[:, 15:-15, 12:-12] = 60.0
    fdata[1] += 5
    rdic = {"data": fdata, "long_name": "reflectivity", "units": "dBz"}
    grid.fields = {"reflectivity": rdic}
    return grid


def make_normal_storm(sigma, mu):
    """
    Make a sample Grid with a gaussian storm target.
    """
    test_grid = make_empty_grid([1, 101, 101], [(1, 1), (-50, 50), (-50, 50)])
    x = test_grid.x["data"]
    y = test_grid.y["data"]
    z = test_grid.z["data"]
    zg, yg, xg = np.meshgrid(z, y, x, indexing="ij")
    r = np.sqrt((xg - mu[0]) ** 2 + (yg - mu[1]) ** 2)
    term1 = 1.0 / (sigma * np.sqrt(2.0 * np.pi))
    term2 = -1.0 * (r**2 / (2.0 * sigma**2))
    data = term1 * np.exp(term2)
    rdic = {"data": data, "long_name": "reflectivity", "units": "dBz"}
    test_grid.fields.update({"reflectivity": rdic})
    return test_grid


def make_empty_spectra_radar(nrays, ngates, npulses_max):
    """
    Return a Spectra Radar object.

    Parameters
    ----------
    nrays : int
        Number of rays in the object.
    ngates : int
        Number of gates per ray.
    npulses_max : int
        Number of pulses in each gate.

    Returns
    -------
    radar : RadarSpectra
        Radar spectra object with an empty spectra field, other parameters are
        set to default values.

    """
    if not _XARRAY_AVAILABLE:
        raise MissingOptionalDependency(
            "Xarray is required to use make_empty_spectra_radar "
            "but is not installed!"
        )

    time_dict = get_metadata("time")
    _range_dict = get_metadata("range")
    latitude_dict = get_metadata("latitude")
    longitude_dict = get_metadata("longitude")
    altitude_dict = get_metadata("altitude")
    sweep_number_dict = get_metadata("sweep_number")
    sweep_mode_dict = get_metadata("sweep_mode")
    fixed_angle_dict = get_metadata("fixed_angle")
    sweep_start_ray_index_dict = get_metadata("sweep_start_ray_index")
    sweep_end_ray_index_dict = get_metadata("sweep_end_ray_index")
    azimuth_dict = get_metadata("azimuth")
    elevation_dict = get_metadata("elevation")

    fields = {}
    scan_type = "vpt"
    c = 299792458
    wavelength = c / 34.830 * 1e-9
    metadata = {"instrument_name": "fake_spectra_radar", "wavelength": wavelength}
    time_dict["units"] = "seconds since 1989-01-01T00:00:01Z"
    time = xr.DataArray(np.arange(nrays, dtype="float32"), attrs=time_dict, dims="time")
    _range = xr.DataArray(
        np.linspace(0, 1000, ngates).astype("float32"), attrs=_range_dict, dims="range"
    )
    latitude = xr.DataArray(np.array([36.5], dtype="float64"), attrs=latitude_dict)
    longitude = xr.DataArray(np.array([-97.5], dtype="float64"), attrs=longitude_dict)
    altitude = xr.DataArray(np.array([200], dtype="float64"), attrs=altitude_dict)

    sweep_number = xr.DataArray(np.arange(1, dtype="int32"), attrs=sweep_number_dict)
    sweep_mode = xr.DataArray(np.array(["spectra"] * 1), attrs=sweep_mode_dict)
    fixed_angle = xr.DataArray(
        np.array([0.75] * 1, dtype="float32"), attrs=fixed_angle_dict
    )
    sweep_start_ray_index = xr.DataArray(
        np.array([0], dtype="int32"), attrs=sweep_start_ray_index_dict
    )
    sweep_end_ray_index = xr.DataArray(
        np.array([len(time.values) - 1], dtype="int32"), attrs=sweep_end_ray_index_dict
    )

    azimuth = xr.DataArray(
        np.arange(nrays, dtype="float32"), attrs=azimuth_dict, dims="time"
    )
    elevation = xr.DataArray(
        np.array([0.75] * nrays, dtype="float32"), attrs=elevation_dict, dims="time"
    )
    npulses_max = np.arange(npulses_max, dtype="float32")
    velocity_bins = xr.DataArray(
        np.linspace(-10.0, 10.0, len(npulses_max)), attrs={}, dims="npulses_max"
    )

    fields = np.zeros((len(time.values), len(_range.values), len(npulses_max)))

    return RadarSpectra(
        time,
        _range,
        fields,
        metadata,
        scan_type,
        latitude,
        longitude,
        altitude,
        sweep_number,
        sweep_mode,
        fixed_angle,
        sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth,
        elevation,
        npulses_max,
        velocity_bins,
        instrument_parameters=None,
    )


def make_target_spectra_radar():
    """
    Return a spectra radar with a target like spectra field.
    """
    if not _XARRAY_AVAILABLE:
        raise MissingOptionalDependency(
            "Xarray is required to use make_target_spectra_radar "
            "but is not installed!"
        )

    radar = make_empty_spectra_radar(10, 20, 50)
    fdata = np.zeros((10, 20, 50), dtype="float32")
    max_value = 10 ** (-10 / 10)
    fdata[:, :, :] = 10 * np.log10(scipy.signal.gaussian(50, std=7) * max_value)
    radar.ds["spectra"].values = fdata
    return radar



"""
General meteorological calculations useful to other modules.
"""

import heapq
import os

import netCDF4
import numpy as np
from pylab import date2num, datestr2num


def nth_smallest(n, iter):
    return heapq.nsmallest(n, iter)[-1]


def get_best_sounding(target, sdir, minl, maxl):
    sondes = os.listdir(sdir)
    sondes.sort()
    offsets = [
        np.abs(datestr2num(s[18:33].replace(".", " ")) - date2num(target))
        for s in sondes
    ]
    cont = True
    n = 1
    while cont:
        test_sonde = sondes[offsets.index(nth_smallest(n, offsets))]
        ncf_obj = netCDF4.Dataset(sdir + test_sonde, "r")
        ncf_min = ncf_obj.variables["alt"][:].min()
        ncf_max = ncf_obj.variables["alt"][:].max()
        if ncf_min < minl and ncf_max > maxl:
            cont = False
            chosen_sonde = test_sonde
        ncf_obj.close()
        n = n + 1
    return chosen_sonde



""" Unit Tests for Py-ART's map/grid_mapper.py module. """

import numpy as np
import pytest
from numpy.testing import assert_almost_equal

import pyart

EXPECTED_CENTER_SLICE = [40, 30, 20, 10, 0, 0, 10, 20, 30, 40]

COMMON_MAP_TO_GRID_ARGS = {
    "grid_shape": (3, 9, 10),
    "grid_limits": ((-400.0, 400.0), (-900.0, 900.0), (-900, 900)),
    "fields": None,
    "gridding_algo": "map_to_grid",
    "roi_func": lambda z, y, x: 30,
}


def test_map_to_grid_filter():

    # simulate a radar with bad gates which reports huge reflectivities
    radar = pyart.testing.make_target_radar()
    radar.fields["reflectivity"]["data"][0:100, 25] = 99999.0

    # without filtering bad gates leaks through
    gatefilter = pyart.filters.GateFilter(radar)
    grids = pyart.map.map_to_grid(
        (radar,), gatefilters=(gatefilter,), **COMMON_MAP_TO_GRID_ARGS
    )
    assert grids["reflectivity"].max() > 41.0

    # with filtering bad gates is supressed
    gatefilter = pyart.filters.GateFilter(radar)
    gatefilter.exclude_above("reflectivity", 41.0)
    grids = pyart.map.map_to_grid(
        (radar,), gatefilters=(gatefilter,), **COMMON_MAP_TO_GRID_ARGS
    )
    assert grids["reflectivity"].max() < 41.0


def test_map_to_grid_non_tuple():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid(radar, **COMMON_MAP_TO_GRID_ARGS)
    center_slice = grids["reflectivity"][1, 4, :]
    assert_almost_equal(center_slice, EXPECTED_CENTER_SLICE)


def test_map_to_grid_default():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid((radar,), **COMMON_MAP_TO_GRID_ARGS)
    center_slice = grids["reflectivity"][1, 4, :]
    assert_almost_equal(center_slice, EXPECTED_CENTER_SLICE)


def test_map_to_grid_cressman():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid(
        (radar,),
        (3, 9, 10),
        ((-400.0, 400.0), (-900.0, 900.0), (-900, 900)),
        roi_func="constant",
        constant_roi=30.0,
        weighting_function="CRESSMAN",
    )
    center_slice = grids["reflectivity"][1, 4, :]
    assert_almost_equal(center_slice, EXPECTED_CENTER_SLICE)


def test_map_to_grid_constant_roi():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid(
        (radar,),
        (3, 9, 10),
        ((-400.0, 400.0), (-900.0, 900.0), (-900, 900)),
        roi_func="constant",
        constant_roi=30.0,
    )
    center_slice = grids["reflectivity"][1, 4, :]
    assert_almost_equal(center_slice, EXPECTED_CENTER_SLICE)


def test_map_to_grid_dist_roi():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid(
        (radar,),
        (3, 9, 10),
        ((-400.0, 400.0), (-900.0, 900.0), (-900, 900)),
        roi_func="dist",
        z_factor=0,
        xy_factor=0,
        min_radius=30.0,
    )
    center_slice = grids["reflectivity"][1, 4, :]
    assert_almost_equal(center_slice, EXPECTED_CENTER_SLICE)


def test_map_to_grid_dist_beam_roi():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid(
        (radar,),
        grid_shape=(3, 9, 10),
        grid_limits=((-400.0, 400.0), (-900.0, 900.0), (-900, 900)),
        fields=["reflectivity"],
        min_radius=30,
        bsp=0.0,
        h_factor=0.0,
    )
    center_slice = grids["reflectivity"][1, 4, :]
    assert_almost_equal(center_slice, EXPECTED_CENTER_SLICE)


def test_map_to_grid_default_two_radars():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid((radar, radar), **COMMON_MAP_TO_GRID_ARGS)
    center_slice = grids["reflectivity"][1, 4, :]
    assert_almost_equal(center_slice, EXPECTED_CENTER_SLICE)


def test_map_to_grid_masked_refl_field():
    radar = pyart.testing.make_target_radar()

    # mask the last gate of the first ray
    fdata = radar.fields["reflectivity"]["data"]
    fdata = np.ma.masked_invalid(fdata)
    fdata.mask = False
    fdata.mask[0, -1] = [True]
    radar.fields["reflectivity"]["data"] = fdata

    grids = pyart.map.map_to_grid((radar,), **COMMON_MAP_TO_GRID_ARGS)
    center_slice = grids["reflectivity"][1, 4, :]
    assert_almost_equal(center_slice, EXPECTED_CENTER_SLICE)


def test_map_to_grid_barnes_filter_masked():

    radar = pyart.testing.make_target_radar()

    # mask and replace with non-sense values a region of radar volume
    fdata = radar.fields["reflectivity"]["data"]
    fdata = np.ma.array(fdata)
    fdata[100:120, 20:40] = -9999.0
    fdata[100:120, 20:40] = np.ma.masked
    radar.fields["reflectivity"]["data"] = fdata

    # check that non-sense values are masked
    assert radar.fields["reflectivity"]["data"].min() > -1.0

    grids = pyart.map.map_to_grid(
        (radar,), weighting_function="Barnes", **COMMON_MAP_TO_GRID_ARGS
    )

    # check that no masked values were included in the interpolation
    assert grids["reflectivity"].min() > -1.0


def test_map_to_grid_no_copy():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid(
        (radar,), copy_field_data=False, **COMMON_MAP_TO_GRID_ARGS
    )
    center_slice = grids["reflectivity"][1, 4, :]
    assert_almost_equal(center_slice, EXPECTED_CENTER_SLICE)


def test_map_to_grid_no_copy_two_radars():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid(
        (radar, radar), copy_field_data=False, **COMMON_MAP_TO_GRID_ARGS
    )
    center_slice = grids["reflectivity"][1, 4, :]
    assert_almost_equal(center_slice, EXPECTED_CENTER_SLICE)


def test_map_to_grid_tiny_grid():
    radar = pyart.testing.make_target_radar()
    grids = pyart.map.map_to_grid(
        (radar,),
        grid_shape=(1, 1, 1),
        grid_limits=((-400.0, 400.0), (-900.0, 900.0), (-900, 900)),
        fields=["reflectivity"],
    )
    assert grids["reflectivity"].shape == (1, 1, 1)
    assert int(grids["reflectivity"][0]) == 40


def test_map_to_grid_errors():
    pytest.raises(
        ValueError,
        pyart.map.map_to_grid,
        (None,),
        (1, 1, 1),
        ((-1, 1), (-1, 1), (-1, 1)),
        weighting_function="foo",
    )
    pytest.raises(
        ValueError,
        pyart.map.map_to_grid,
        (None,),
        (1, 1, 1),
        ((-1, 1), (-1, 1), (-1, 1)),
        algorithm="foo",
    )
    radar = pyart.testing.make_target_radar()
    pytest.raises(
        ValueError,
        pyart.map.map_to_grid,
        (radar,),
        (1, 1, 1),
        ((-1, 1), (-1, 1), (-1, 1)),
        roi_func="foo",
    )


def test_grid_from_radars_errors():
    pytest.raises(
        ValueError,
        pyart.map.grid_from_radars,
        (None,),
        (1, 1, 1),
        ((-1, 1), (-1, 1), (-1, 1)),
        gridding_algo="foo",
    )


def test_grid_from_radars():
    radar = pyart.testing.make_target_radar()
    grid = pyart.map.grid_from_radars((radar,), **COMMON_MAP_TO_GRID_ARGS)

    # check field data
    center_slice = grid.fields["reflectivity"]["data"][1, 4, :]
    assert_almost_equal(center_slice, EXPECTED_CENTER_SLICE)

    # check other Grid object attributes
    assert "ROI" in grid.fields
    assert np.all(grid.fields["ROI"]["data"] == 30.0)
    assert_almost_equal(grid.x["data"], np.linspace(-900, 900, 10))
    assert_almost_equal(grid.y["data"], np.linspace(-900, 900, 9))
    assert_almost_equal(grid.z["data"], np.linspace(-400, 400, 3))

    # check that grid.radar_ attributes set correctly
    assert isinstance(grid.radar_latitude, dict)
    assert_almost_equal(grid.radar_latitude["data"][0], 36.5)

    assert isinstance(grid.radar_longitude, dict)
    assert_almost_equal(grid.radar_longitude["data"][0], -97.5)

    assert isinstance(grid.radar_altitude, dict)
    assert_almost_equal(grid.radar_altitude["data"][0], 200)

    assert isinstance(grid.radar_time, dict)
    assert_almost_equal(grid.radar_time["data"][0], 0.0)
    assert grid.radar_time["units"] == radar.time["units"]

    assert isinstance(grid.radar_name, dict)
    assert grid.radar_name["data"][0] == "fake_radar"

    assert grid.nradar == 1


def test_unify_times_for_radars():
    radar1 = pyart.testing.make_target_radar()
    radar2 = pyart.testing.make_target_radar()
    radar2.time["units"] = "seconds since 1989-01-01T00:00:02Z"
    radars = (radar1, radar2)
    times, units = pyart.map.grid_mapper._unify_times_for_radars(radars)
    assert_almost_equal(times[0], 0)
    assert_almost_equal(times[1], 1)
    assert units == "seconds since 1989-01-01T00:00:01Z"


def test_grid_from_radars_non_tuple():
    radar = pyart.testing.make_target_radar()
    grid = pyart.map.grid_from_radars(radar, **COMMON_MAP_TO_GRID_ARGS)

    # check field data
    center_slice = grid.fields["reflectivity"]["data"][1, 4, :]
    assert_almost_equal(center_slice, EXPECTED_CENTER_SLICE)

    # check other Grid object attributes
    assert "ROI" in grid.fields
    assert np.all(grid.fields["ROI"]["data"] == 30.0)
    assert_almost_equal(grid.x["data"], np.linspace(-900, 900, 10))
    assert_almost_equal(grid.y["data"], np.linspace(-900, 900, 9))
    assert_almost_equal(grid.z["data"], np.linspace(-400, 400, 3))


def test_grid_from_radars_grid_origin():
    radar = pyart.testing.make_target_radar()
    radar.metadata.pop("instrument_name")
    grid = pyart.map.grid_from_radars(
        (radar,),
        grid_origin=(36.4, -97.6),
        grid_origin_alt=200,
        **COMMON_MAP_TO_GRID_ARGS
    )
    assert_almost_equal(grid.origin_latitude["data"][0], 36.4, 1)
    assert_almost_equal(grid.origin_longitude["data"], -97.6, 1)


def test_example_roi_funcs():
    assert pyart.map.example_roi_func_constant(0, 0, 0) == 500.0
    assert pyart.map.example_roi_func_dist(0, 0, 0) == 500.0
    assert pyart.map.example_roi_func_dist_beam(0, 0, 0) == 500.0



"""
Utilities for reading EDGE NetCDF files.

"""

import datetime

import netCDF4
import numpy as np

from ..config import FileMetadata, get_fillvalue
from ..core.radar import Radar
from ..io.common import _test_arguments, make_time_unit_str


def read_edge_netcdf(filename, **kwargs):
    """
    Read a EDGE NetCDF file.

    Parameters
    ----------
    filename : str
        Name of EDGE NetCDF file to read data from.

    Returns
    -------
    radar : Radar
        Radar object.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    filemetadata = FileMetadata("edge_netcdf")

    # Open netCDF4 file
    dset = netCDF4.Dataset(filename)
    nrays = len(dset.dimensions["Azimuth"])
    nbins = len(dset.dimensions["Gate"])

    # latitude, longitude and altitude
    latitude = filemetadata("latitude")
    longitude = filemetadata("longitude")
    altitude = filemetadata("altitude")
    latitude["data"] = np.array([dset.Latitude], "float64")
    longitude["data"] = np.array([dset.Longitude], "float64")
    altitude["data"] = np.array([dset.Height], "float64")

    # metadata
    metadata = filemetadata("metadata")
    metadata_mapping = {
        "vcp-value": "vcp",
        "radarName-value": "radar_name",
        "ConversionPlugin": "conversion_software",
    }
    for netcdf_attr, metadata_key in metadata_mapping.items():
        if netcdf_attr in dset.ncattrs():
            metadata[metadata_key] = dset.getncattr(netcdf_attr)

    # sweep_start_ray_index, sweep_end_ray_index
    sweep_start_ray_index = filemetadata("sweep_start_ray_index")
    sweep_end_ray_index = filemetadata("sweep_end_ray_index")
    sweep_start_ray_index["data"] = np.array([0], dtype="int32")
    sweep_end_ray_index["data"] = np.array([nrays - 1], dtype="int32")

    # sweep number
    sweep_number = filemetadata("sweep_number")
    sweep_number["data"] = np.array([0], dtype="int32")

    # sweep_type
    scan_type = "ppi"

    # sweep_mode, fixed_angle
    sweep_mode = filemetadata("sweep_mode")
    fixed_angle = filemetadata("fixed_angle")
    sweep_mode["data"] = np.array(1 * ["azimuth_surveillance"])
    fixed_angle["data"] = np.array([dset.Elevation], dtype="float32")

    # time
    time = filemetadata("time")
    start_time = datetime.datetime.utcfromtimestamp(dset.Time)
    time["units"] = make_time_unit_str(start_time)
    time["data"] = np.zeros((nrays,), dtype="float64")

    # range
    _range = filemetadata("range")
    step = float(dset.getncattr("MaximumRange-value")) / nbins * 1000.0
    _range["data"] = np.arange(nbins, dtype="float32") * step + step / 2
    _range["meters_to_center_of_first_gate"] = step / 2.0
    _range["meters_between_gates"] = step

    # elevation
    elevation = filemetadata("elevation")
    elevation_angle = dset.Elevation
    elevation["data"] = np.ones((nrays,), dtype="float32") * elevation_angle

    # azimuth
    azimuth = filemetadata("azimuth")
    azimuth["data"] = dset.variables["Azimuth"][:]

    # fields
    field_name = dset.TypeName

    field_data = np.ma.array(dset.variables[field_name][:])
    if "MissingData" in dset.ncattrs():
        field_data[field_data == dset.MissingData] = np.ma.masked
    if "RangeFolded" in dset.ncattrs():
        field_data[field_data == dset.RangeFolded] = np.ma.masked

    fields = {field_name: filemetadata(field_name)}
    fields[field_name]["data"] = field_data
    fields[field_name]["units"] = dset.variables[field_name].Units
    fields[field_name]["_FillValue"] = get_fillvalue()

    # instrument_parameters
    instrument_parameters = {}

    if "PRF-value" in dset.ncattrs():
        dic = filemetadata("prt")
        prt = 1.0 / float(dset.getncattr("PRF-value"))
        dic["data"] = np.ones((nrays,), dtype="float32") * prt
        instrument_parameters["prt"] = dic

    if "PulseWidth-value" in dset.ncattrs():
        dic = filemetadata("pulse_width")
        pulse_width = dset.getncattr("PulseWidth-value") * 1.0e-6
        dic["data"] = np.ones((nrays,), dtype="float32") * pulse_width
        instrument_parameters["pulse_width"] = dic

    if "NyquistVelocity-value" in dset.ncattrs():
        dic = filemetadata("nyquist_velocity")
        nyquist_velocity = float(dset.getncattr("NyquistVelocity-value"))
        dic["data"] = np.ones((nrays,), dtype="float32") * nyquist_velocity
        instrument_parameters["nyquist_velocity"] = dic

    if "Beamwidth" in dset.variables:
        dic = filemetadata("radar_beam_width_h")
        dic["data"] = dset.variables["Beamwidth"][:]
        instrument_parameters["radar_beam_width_h"] = dic

    dset.close()

    return Radar(
        time,
        _range,
        fields,
        metadata,
        scan_type,
        latitude,
        longitude,
        altitude,
        sweep_number,
        sweep_mode,
        fixed_angle,
        sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth,
        elevation,
        instrument_parameters=instrument_parameters,
    )



#! /usr/bin/env python

import netCDF4

dset = netCDF4.Dataset("sgpinterpolatedsondeC1.c1.20110510.000000.cdf")
dvars = dset.variables

out = netCDF4.Dataset("example_interpolatedsonde.cdf", mode="w")

out.createDimension("time", None)
out.createDimension("height", 316)

time = out.createVariable("time", dvars["time"].dtype, ("time"))
height = out.createVariable("height", dvars["height"].dtype, ("height"))
wspd = out.createVariable("wspd", dvars["wspd"].dtype, ("time", "height"))
wdir = out.createVariable("wdir", dvars["wdir"].dtype, ("time", "height"))

time[:] = dvars["time"][685:695]
time.units = dvars["time"].units
height[:] = dvars["height"][:]
wspd[:] = dvars["wspd"][685:695, :]
wdir[:] = dvars["wdir"][685:695, :]

out.close()



"""
Functions for converting date and time between various forms.

"""

try:
    from cftime import date2num, num2date
except ImportError:
    from netCDF4 import date2num, num2date

EPOCH_UNITS = "seconds since 1970-01-01T00:00:00Z"


def datetime_from_radar(radar, epoch=False, **kwargs):
    """Return a datetime for the first ray in a Radar."""
    if epoch:
        dtrad = num2date(radar.time["data"][0], radar.time["units"])
        epnum = date2num(dtrad, EPOCH_UNITS)
        return num2date(epnum, EPOCH_UNITS, **kwargs)
    else:
        return num2date(radar.time["data"][0], radar.time["units"], **kwargs)


def datetimes_from_radar(radar, epoch=False, **kwargs):
    """Return an array of datetimes for the rays in a Radar."""
    if epoch:
        dtrad = num2date(radar.time["data"][:], radar.time["units"])
        epnum = date2num(dtrad, EPOCH_UNITS)
        return num2date(epnum, EPOCH_UNITS, **kwargs)
    else:
        return num2date(radar.time["data"][:], radar.time["units"], **kwargs)


def datetime_from_dataset(dataset, epoch=False, **kwargs):
    """Return a datetime for the first time in a netCDF Dataset."""
    if epoch:
        dtdata = num2date(dataset.variables["time"][0], dataset.variables["time"].units)
        epnum = date2num(dtdata, EPOCH_UNITS)
        return num2date(epnum, EPOCH_UNITS, **kwargs)
    else:
        return num2date(
            dataset.variables["time"][0], dataset.variables["time"].units, **kwargs
        )


def datetimes_from_dataset(dataset, epoch=False, **kwargs):
    """Return an array of datetimes for the times in a netCDF Dataset."""
    if epoch:
        dtdata = num2date(dataset.variables["time"][:], dataset.variables["time"].units)
        epnum = date2num(dtdata, EPOCH_UNITS)
        return num2date(epnum, EPOCH_UNITS, **kwargs)
    else:
        return num2date(
            dataset.variables["time"][:], dataset.variables["time"].units, **kwargs
        )


def datetime_from_grid(grid, epoch=False, **kwargs):
    """Return a datetime for the volume start in a Grid."""
    if epoch:
        dtrad = num2date(grid.time["data"][0], grid.time["units"])
        epnum = date2num(dtrad, EPOCH_UNITS)
        return num2date(epnum, EPOCH_UNITS, **kwargs)
    else:
        return num2date(grid.time["data"][0], grid.time["units"], **kwargs)



""" Unit Tests for Py-ART's graph/gridmapdisplay.py module. """
# execute this script to create figure_gridmapdisplay_*.png files.

# TODO use matplotlib's @image_comparison decorator to compare to file
# in baseline_images directory. Currently this test only determines if files
# can be created, not that they are correct.

import matplotlib.pyplot as plt
import pytest

import pyart


@pytest.mark.skipif(
    not pyart.graph.gridmapdisplay._CARTOPY_AVAILABLE, reason="Cartopy is not installed"
)
def test_gridmapdisplay_simple(outfile=None):
    # test basic GridMapDisplat functionally.
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplay(grid)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_grid("reflectivity", vmin=-5, vmax=35, ax=ax)
    if outfile:
        fig.savefig(outfile)


@pytest.mark.skipif(
    not pyart.graph.gridmapdisplay._CARTOPY_AVAILABLE, reason="Cartopy is not installed"
)
def test_gridmapdisplay_imshow(outfile=None):
    # test basic GridMapDisplay functionality.
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplay(grid)
    fig = plt.figure()
    ax = fig.add_subplot(111)
    display.plot_grid("reflectivity", imshow=True, vmin=-5, vmax=35, ax=ax)
    if outfile:
        fig.savefig(outfile)


@pytest.mark.skipif(
    not pyart.graph.gridmapdisplay._CARTOPY_AVAILABLE, reason="Cartopy is not installed"
)
def test_gridmapdisplay_fancy(outfile=None):
    import cartopy.crs as ccrs

    # test a bunch of GridMapDisplay functionaliy
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplay(grid)
    fig = plt.figure(1, figsize=[25.0, 17.0])
    projection = ccrs.Mercator()

    ax1 = plt.subplot(331, projection=projection)
    display.plot_grid(
        "reflectivity",
        vmin=-5.0,
        vmax=35.0,
        ax=ax1,
        mask_outside=True,
        axislabels=("foo", "bar"),
        axislabels_flag=True,
        title="Special title",
    )
    display.plot_crosshairs(color="b")

    ax2 = plt.subplot(332, projection=projection)
    display.plot_grid("reflectivity", ax=ax2, axislabels_flag=True)

    ax3 = plt.subplot(333)
    display.plot_colorbar(ax=ax3)

    ax4 = plt.subplot(334)
    display.plot_latitude_slice("reflectivity", ax=ax4, mask_outside=True)

    ax5 = plt.subplot(335)
    display.plot_latitude_slice("reflectivity", ax=ax5, title="Lat title")

    ax6 = plt.subplot(336)
    grid.fields["reflectivity"]["valid_min"] = 0
    grid.fields["reflectivity"]["valid_max"] = 30
    display.plot_longitude_slice("reflectivity", ax=ax6, mask_outside=True)

    ax7 = plt.subplot(337)
    display.plot_longitude_slice("reflectivity", ax=ax7, title="Lon title")

    ax8 = plt.subplot(338)
    del display.grid.fields["reflectivity"]["long_name"]
    display.plot_colorbar(ax=ax8)

    if outfile:
        fig.savefig(outfile)


@pytest.mark.skipif(
    not pyart.graph.gridmapdisplay._CARTOPY_AVAILABLE, reason="Cartopy is not installed"
)
def test_generate_filename():
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplay(grid)
    filename = display.generate_filename("reflectivity", 0)
    assert isinstance(filename, str)


@pytest.mark.skipif(
    not pyart.graph.gridmapdisplay._CARTOPY_AVAILABLE, reason="Cartopy is not installed"
)
def test_generate_titles():
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplay(grid)

    title = display.generate_longitudinal_level_title("reflectivity", 0)
    assert isinstance(title, str)

    title = display.generate_latitudinal_level_title("reflectivity", 0)
    assert isinstance(title, str)


@pytest.mark.skipif(
    not pyart.graph.gridmapdisplay._CARTOPY_AVAILABLE, reason="Cartopy is not installed"
)
def test_error_raising():
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplay(grid)

    # no mappable
    pytest.raises(ValueError, display.plot_colorbar)

    # no field
    display.mappables.append(None)  # mock the mappable
    pytest.raises(ValueError, display.plot_colorbar)


@pytest.mark.mpl_image_compare(tolerance=30)
@pytest.mark.skipif(
    not pyart.graph.gridmapdisplay._METPY_AVAILABLE, reason="MetPy is not installed"
)
def test_gridmapdisplay_cross_section(outfile=None):
    # test basic GridMapDisplay cross section functionality.
    start = (34.8, -98.75)
    end = (38.6, -96.45)
    fig = plt.figure()
    grid = pyart.testing.make_target_grid()
    display = pyart.graph.GridMapDisplay(grid)
    display.plot_cross_section(
        "reflectivity", start, end, x_axis="lat", vmin=-5, vmax=35
    )
    try:
        return fig
    finally:
        plt.close(fig)


if __name__ == "__main__":
    test_gridmapdisplay_simple("figure_grid_mapdisplay_simple.png")
    test_gridmapdisplay_fancy("figure_gridmapdisplay_fancy.png")



"""
=========================================
Calculate and Plot VAD profile
=========================================

Calculates a VAD and plots a vertical profile of wind
"""

# Author: Daniel Wolfensberger (daniel.wolfensberger@meteoswiss.ch)
# License: BSD 3 clause

import matplotlib.pyplot as plt
import numpy as np
from open_radar_data import DATASETS

import pyart

# Read in a sample file
filename = DATASETS.fetch("MLA2119412050U.nc")
radar = pyart.io.read_cfradial(filename)

# Loop on all sweeps and compute VAD
zlevels = np.arange(100, 5000, 100)  # height above radar
u_allsweeps = []
v_allsweeps = []

for idx in range(radar.nsweeps):
    radar_1sweep = radar.extract_sweeps([idx])
    vad = pyart.retrieve.vad_browning(
        radar_1sweep, "corrected_velocity", z_want=zlevels
    )
    u_allsweeps.append(vad.u_wind)
    v_allsweeps.append(vad.v_wind)

# Average U and V over all sweeps and compute magnitude and angle
u_avg = np.nanmean(np.array(u_allsweeps), axis=0)
v_avg = np.nanmean(np.array(v_allsweeps), axis=0)
orientation = np.rad2deg(np.arctan2(-u_avg, -v_avg)) % 360
speed = np.sqrt(u_avg**2 + v_avg**2)

# Display vertical profile of wind
fig, ax = plt.subplots(1, 2, sharey=True)
ax[0].plot(speed * 2, zlevels + radar.altitude["data"])
ax[1].plot(orientation, zlevels + radar.altitude["data"])
ax[0].set_xlabel("Wind speed [m/s]")
ax[1].set_xlabel("Wind direction [deg]")
ax[0].set_ylabel("Altitude [m]")
fig.suptitle("Wind profile obtained from VAD")



""" Unit Tests for Py-ART's io/sigmet.py module. """

import numpy as np
import pytest
from numpy.ma.core import MaskedArray
from numpy.testing import assert_almost_equal

import pyart

############################################
# read_rsl tests (verify radar attributes) #
############################################

# read in the sample file and create a a Radar object
radar = pyart.io.read_sigmet(pyart.testing.SIGMET_PPI_FILE)


# time attribute
def test_time():
    assert "comment" in radar.time.keys()
    assert "long_name" in radar.time.keys()
    assert "standard_name" in radar.time.keys()
    assert "units" in radar.time.keys()
    assert "calendar" in radar.time.keys()
    assert "data" in radar.time.keys()
    assert radar.time["units"] == "seconds since 2011-05-20T10:54:08Z"
    assert radar.time["data"].shape == (20,)
    assert_almost_equal(radar.time["data"][1], 1, 0)


# range attribute
def test_range():
    assert "long_name" in radar.range
    assert "standard_name" in radar.range
    assert "meters_to_center_of_first_gate" in radar.range
    assert "meters_between_gates" in radar.range
    assert "units" in radar.range
    assert "data" in radar.range
    assert "spacing_is_constant" in radar.range
    assert radar.range["data"].shape == (25,)
    assert_almost_equal(radar.range["data"][1], 62, 0)


# fields attribute is tested later


# metadata attribute
def test_metadata():
    assert "instrument_name" in radar.metadata
    assert "source" in radar.metadata


# scan_type attribute
def test_scan_type():
    assert radar.scan_type == "ppi"


# latitude attribute
def test_latitude():
    assert "data" in radar.latitude
    assert "standard_name" in radar.latitude
    assert "units" in radar.latitude
    assert radar.latitude["data"].shape == (1,)
    assert_almost_equal(radar.latitude["data"], 36, 0)


# longitude attribute
def test_longitude():
    assert "data" in radar.longitude
    assert "standard_name" in radar.longitude
    assert "units" in radar.longitude
    assert radar.longitude["data"].shape == (1,)
    assert_almost_equal(radar.longitude["data"], -98, 0)


# altitude attribute
def test_altitude():
    assert "data" in radar.altitude
    assert "standard_name" in radar.altitude
    assert "units" in radar.altitude
    assert "positive" in radar.altitude
    assert radar.altitude["data"].shape == (1,)
    assert_almost_equal(radar.altitude["data"], 214, 0)


# altitude_agl attribute
def test_altitude_agl():
    assert radar.altitude_agl is None


# sweep_number attribute
def test_sweep_number():
    assert "standard_name" in radar.sweep_number
    assert np.all(radar.sweep_number["data"] == range(1))


# sweep_mode attribute
def test_sweep_mode():
    assert "standard_name" in radar.sweep_mode
    assert radar.sweep_mode["data"].shape == (1,)
    assert radar.sweep_mode["data"].dtype.char == "S"
    assert np.all(radar.sweep_mode["data"] == [b"azimuth_surveillance"])


# fixed_angle attribute
def test_fixed_angle():
    assert "standard_name" in radar.fixed_angle
    assert "units" in radar.fixed_angle
    assert radar.fixed_angle["data"].shape == (1,)
    assert_almost_equal(radar.fixed_angle["data"][0], 0.50, 2)


# sweep_start_ray_index attribute
def test_sweep_start_ray_index():
    assert "long_name" in radar.sweep_start_ray_index
    assert radar.sweep_start_ray_index["data"].shape == (1,)
    assert_almost_equal(radar.sweep_start_ray_index["data"][0], 0, 0)


# sweep_end_ray_index attribute
def test_sweep_end_ray_index():
    assert "long_name" in radar.sweep_end_ray_index
    assert radar.sweep_end_ray_index["data"].shape == (1,)
    assert_almost_equal(radar.sweep_end_ray_index["data"][0], 19, 0)


# target_scan_rate attribute
def test_target_scan_rate():
    assert radar.target_scan_rate is None


# azimuth attribute
def test_azimuth():
    assert "standard_name" in radar.azimuth
    assert "long_name" in radar.azimuth
    assert "units" in radar.azimuth
    assert "axis" in radar.azimuth
    assert_almost_equal(radar.azimuth["data"][0], 0, 0)
    assert_almost_equal(radar.azimuth["data"][10], 180, 0)


# elevation attribute
def test_elevation():
    assert "standard_name" in radar.elevation
    assert "long_name" in radar.azimuth
    assert "units" in radar.elevation
    assert "axis" in radar.elevation
    assert radar.elevation["data"].shape == (20,)
    assert_almost_equal(radar.elevation["data"][0], 0.50, 2)


# scan_rate attribute
def test_scan_rate():
    assert radar.scan_rate is None


# antenna_transition attribute
def test_antenna_transition():
    assert radar.antenna_transition is None


# instrument_parameters attribute
keys = ["prt", "unambiguous_range", "prt_mode", "nyquist_velocity", "pulse_width"]


@pytest.mark.parametrize("keys", keys)
def test_instument_parameters(keys):
    # instrument_parameter sub-convention
    description = "instrument_parameters: %s" % keys
    check_instrument_parameter.description = description
    check_instrument_parameter(keys)


def check_instrument_parameter(param):
    assert param in radar.instrument_parameters
    param_dic = radar.instrument_parameters[param]
    assert param_dic["meta_group"] == "instrument_parameters"


def test_prt_mode_dtype():
    assert radar.instrument_parameters["prt_mode"]["data"].dtype.char == "S"


# radar_parameters attribute
@pytest.mark.parametrize("keys", ["radar_beam_width_h", "radar_beam_width_v"])
def test_radar_parameters(keys):
    # radar_parameter sub-convention
    description = "radar_parameters: %s" % keys
    check_radar_parameter.description = description
    check_radar_parameter(keys)


def check_radar_parameter(param):
    assert param in radar.instrument_parameters
    param_dic = radar.instrument_parameters[param]
    assert param_dic["meta_group"] == "radar_parameters"


# radar_calibration attribute
def test_radar_calibration():
    assert radar.radar_calibration is None


# ngates attribute
def test_ngates():
    assert radar.ngates == 25


# nrays attribute
def test_nrays():
    assert radar.nrays == 20


# nsweeps attribute
def test_nsweeps():
    assert radar.nsweeps == 1


####################
# fields attribute #
####################


@pytest.mark.parametrize(
    "field",
    [
        "reflectivity",
    ],
)
def test_field_dics(field):
    description = "field : %s, dictionary" % field
    check_field_dic.description = description
    check_field_dic(field)


def check_field_dic(field):
    """Check that the required keys are present in a field dictionary."""
    assert "standard_name" in radar.fields[field]
    assert "units" in radar.fields[field]
    assert "_FillValue" in radar.fields[field]
    assert "coordinates" in radar.fields[field]


@pytest.mark.parametrize(
    "field",
    [
        "reflectivity",
    ],
)
def test_field_shapes(field):
    description = "field : %s, shape" % field
    check_field_shape.description = description
    check_field_shape(field)


def check_field_shape(field):
    assert radar.fields[field]["data"].shape == (20, 25)


fields = {
    "reflectivity": MaskedArray,
}


@pytest.mark.parametrize("field, field_type", fields.items(), ids=list(fields.keys()))
def test_field_types(field, field_type):
    description = "field : %s, type" % field
    check_field_type.description = description
    check_field_type(field, field_type)


def check_field_type(field, field_type):
    assert type(radar.fields[field]["data"]) is field_type


fields = {
    "reflectivity": 0.0,
}


@pytest.mark.parametrize("field, field_value", fields.items(), ids=list(fields.keys()))
def test_field_first_points(field, field_value):
    # these values can be found using:
    # [round(radar.fields[f]['data'][0,0]) for f in radar.fields]
    description = "field : %s, first point" % field
    check_field_first_point.description = description
    check_field_first_point(field, field_value)


def check_field_first_point(field, value):
    assert_almost_equal(radar.fields[field]["data"][0, 0], value, 0)


def test_nonuniform_bin_masking():
    # Check that in non-collected gates are masked for rays in which the
    # number of bins is smaller than the number of bins in the product header.
    assert radar.fields["reflectivity"]["data"][19, 14] == 0.0
    assert radar.fields["reflectivity"]["data"][19, 15] is np.ma.masked
    assert radar.fields["reflectivity"]["data"][19, 16] is np.ma.masked
    assert radar.fields["reflectivity"]["data"][19, 17] is np.ma.masked
    assert radar.fields["reflectivity"]["data"][19, 18] is np.ma.masked
    assert radar.fields["reflectivity"]["data"][19, 19] is np.ma.masked
    assert radar.fields["reflectivity"]["data"][19, 20] is np.ma.masked
    assert radar.fields["reflectivity"]["data"][19, 21] is np.ma.masked
    assert radar.fields["reflectivity"]["data"][19, 22] is np.ma.masked
    assert radar.fields["reflectivity"]["data"][19, 23] is np.ma.masked
    assert radar.fields["reflectivity"]["data"][19, 24] is np.ma.masked


#############
# RHI tests #
#############

RADAR_RHI = pyart.io.read_sigmet(pyart.testing.SIGMET_RHI_FILE)


# nsweeps attribute
def test_rhi_nsweeps():
    assert RADAR_RHI.nsweeps == 1


# sweep_number attribute
def test_rhi_sweep_number():
    assert "standard_name" in RADAR_RHI.sweep_number
    assert np.all(RADAR_RHI.sweep_number["data"] == range(1))


# sweep_mode attribute
def test_rhi_sweep_mode():
    assert "standard_name" in RADAR_RHI.sweep_mode
    assert RADAR_RHI.sweep_mode["data"].shape == (1,)
    assert np.all(RADAR_RHI.sweep_mode["data"] == [b"rhi"])


# fixed_angle attribute
def test_rhi_fixed_angle():
    assert "standard_name" in RADAR_RHI.fixed_angle
    assert "units" in RADAR_RHI.fixed_angle
    assert RADAR_RHI.fixed_angle["data"].shape == (1,)
    assert_almost_equal(RADAR_RHI.fixed_angle["data"][0], 37.30, 2)


# sweep_start_ray_index attribute
def test_rhi_sweep_start_ray_index():
    assert "long_name" in RADAR_RHI.sweep_start_ray_index
    assert RADAR_RHI.sweep_start_ray_index["data"].shape == (1,)
    assert_almost_equal(RADAR_RHI.sweep_start_ray_index["data"][0], 0, 0)


# sweep_end_ray_index attribute
def test_rhi_sweep_end_ray_index():
    assert "long_name" in RADAR_RHI.sweep_end_ray_index
    assert RADAR_RHI.sweep_end_ray_index["data"].shape == (1,)
    assert_almost_equal(RADAR_RHI.sweep_end_ray_index["data"][0], 19, 0)


# azimuth attribute
def test_rhi_azimuth():
    assert "standard_name" in RADAR_RHI.azimuth
    assert "long_name" in RADAR_RHI.azimuth
    assert "units" in RADAR_RHI.azimuth
    assert "axis" in RADAR_RHI.azimuth
    assert_almost_equal(RADAR_RHI.azimuth["data"][0], 37, 0)
    assert_almost_equal(RADAR_RHI.azimuth["data"][10], 37, 0)


# elevation attribute
def test_rhi_elevation():
    assert "standard_name" in RADAR_RHI.elevation
    assert "long_name" in RADAR_RHI.azimuth
    assert "units" in RADAR_RHI.elevation
    assert "axis" in RADAR_RHI.elevation
    assert RADAR_RHI.elevation["data"].shape == (20,)
    assert_almost_equal(RADAR_RHI.elevation["data"][1], 9, 2)


#########
# 1byte #
#########


# Test 1byte sigmet format. github issue #299
def test_1byte_datatype():
    data_type = 1  # Sigmet type for for a DBT field
    data = np.ones((2, 2), dtype=np.int16) * 257
    nbins = np.ones((2,), dtype=np.int16) * 2
    result = pyart.io._sigmetfile.convert_sigmet_data(data_type, data, nbins)
    assert np.all(result == -31.5)
    assert result.shape == (2, 2)



"""
Routines for reading GCPEX D3R files.

"""

import datetime

import netCDF4
import numpy as np

from ..config import FileMetadata
from ..core.radar import Radar
from ..io.common import _test_arguments, make_time_unit_str

D3R_FIELD_NAMES = {
    # corrected reflectivity, horizontal
    "Reflectivity": "reflectivity",
    # corrected reflectivity, vertical
    "DBZV": "reflectivity",
    # differential reflectivity
    "DifferentialReflectivity": "differential_reflectivity",
    "CrossPolCorrelation": "cross_correlation_ratio",
    "ClutterPowerH": "clutter_power_h",
    "ClutterPowerV": "clutter_power_v",
    "DifferentialPhase": "differential_phase",
    "KDP": "specific_differential_phase",
    "NormalizedCoherentPower": "normalized_coherent_power",
    "Signal+Clutter_toNoise_H": "signal_to_noise_ratio",
    "Velocity": "velocity",
    "SpectralWidth": "spectrum_width",
    "SignalPower_H": "signal_power_h",
}


def read_d3r_gcpex_nc(
    filename,
    field_names=None,
    additional_metadata=None,
    file_field_names=False,
    exclude_fields=None,
    include_fields=None,
    read_altitude_from_nc=False,
    **kwargs
):
    """
    Read a D3R GCPEX netCDF file.

    Parameters
    ----------
    filename : str
        Name of the ODIM_H5 file to read.
    field_names : dict, optional
        Dictionary mapping ODIM_H5 field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.
    read_altitude_from_nc : bool, optional
        True if you want the altitude value to be read from the provider netCDF file.
        False will default to the value np.array([295.], dtype='float64')

    Returns
    -------
    radar : Radar
        Radar object containing data from ODIM_H5 file.

    """
    # TODO before moving to pyart.io
    # * unit test
    # * add default field mapping, etc to default config
    # * auto-detect file type with pyart.io.read function
    # * instrument parameters
    # * add additional checks for HOW attributes
    # * support for other objects (SCAN, XSEC)

    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    if field_names is None:
        field_names = D3R_FIELD_NAMES
    filemetadata = FileMetadata(
        "cfradial",
        field_names,
        additional_metadata,
        file_field_names,
        exclude_fields,
        include_fields,
    )

    # read the data
    ncobj = netCDF4.Dataset(filename)
    ncvars = ncobj.variables

    # One sweep per file
    nsweeps = 1

    # latitude, longitude and altitude
    latitude = filemetadata("latitude")
    longitude = filemetadata("longitude")
    altitude = filemetadata("altitude")

    latitude["data"] = np.array([ncobj.Latitude], dtype="float64")
    longitude["data"] = np.array([ncobj.Longitude], dtype="float64")
    altitude_list = [ncobj.Altitude] if read_altitude_from_nc else [295.0]
    altitude["data"] = np.array(altitude_list, dtype="float64")

    # metadata
    metadata = filemetadata("metadata")
    metadata["source"] = "Colorado State EE - chandrasekar"
    metadata["original_container"] = "D3R_gcpex_nc"
    metadata["nc_conventions"] = ncobj.NetCDFRevision

    metadata["version"] = ncobj.NetCDFRevision
    metadata["source"] = "Chandra"

    metadata["system"] = ncobj.RadarName
    metadata["software"] = ncobj.NetCDFRevision
    metadata["sw_version"] = ncobj.NetCDFRevision

    # sweep_start_ray_index, sweep_end_ray_index
    sweep_start_ray_index = filemetadata("sweep_start_ray_index")
    sweep_end_ray_index = filemetadata("sweep_end_ray_index")

    rays_per_sweep = np.shape(ncvars["Azimuth"][:])
    ssri = np.cumsum(np.append([0], rays_per_sweep[:-1])).astype("int32")
    seri = np.cumsum(rays_per_sweep).astype("int32") - 1
    sweep_start_ray_index["data"] = ssri
    sweep_end_ray_index["data"] = seri

    # sweep_number
    sweep_number = filemetadata("sweep_number")
    sweep_number["data"] = np.arange(nsweeps, dtype="int32")

    # sweep_mode
    sweep_mode = filemetadata("sweep_mode")
    sweep_mode["data"] = np.array(nsweeps * ["azimuth_surveillance"])

    # scan_type
    if ncobj.ScanType == 2:
        scan_type = "ppi"
    else:
        scan_type = "rhi"

    # fixed_angle
    fixed_angle = filemetadata("fixed_angle")
    if ncobj.ScanType == 2:
        sweep_el = ncvars["Elevation"][0]
    else:
        sweep_el = ncvars["Azimuth"][0]
    fixed_angle["data"] = np.array([sweep_el], dtype="float32")

    # elevation
    elevation = filemetadata("elevation")
    elevation["data"] = ncvars["Elevation"]

    # range
    _range = filemetadata("range")
    # check that the gate spacing is constant between sweeps
    rstart = ncvars["StartRange"][:]
    if any(rstart != rstart[0]):
        raise ValueError("range start changes between sweeps")
    rscale = ncvars["GateWidth"][:] / 1000.0
    if any(rscale != rscale[0]):
        raise ValueError("range scale changes between sweeps")

    nbins = ncobj.NumGates
    _range["data"] = np.arange(nbins, dtype="float32") * rscale[0] + rstart[0] * 1000.0
    _range["meters_to_center_of_first_gate"] = rstart[0]
    _range["meters_between_gates"] = float(rscale[0])

    # azimuth
    azimuth = filemetadata("azimuth")
    azimuth["data"] = ncvars["Azimuth"][:]

    # time
    _time = filemetadata("time")
    start_time = datetime.datetime.utcfromtimestamp(ncobj.Time)
    _time["units"] = make_time_unit_str(start_time)
    _time["data"] = (ncvars["Time"] - ncobj.Time).astype("float32")

    # fields
    # all variables with dimensions of 'Radial', 'Gate' are fields
    keys = [k for k, v in ncvars.items() if v.dimensions == ("Radial", "Gate")]

    fields = {}
    for key in keys:
        field_name = filemetadata.get_field_name(key)
        if field_name is None:
            if exclude_fields is not None and key in exclude_fields:
                continue
            if include_fields is not None:
                if key not in include_fields:
                    continue
            field_name = key
        fields[field_name] = _ncvar_to_dict(ncvars[key])

    # instrument_parameters
    instrument_parameters = None

    return Radar(
        _time,
        _range,
        fields,
        metadata,
        scan_type,
        latitude,
        longitude,
        altitude,
        sweep_number,
        sweep_mode,
        fixed_angle,
        sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth,
        elevation,
        instrument_parameters=instrument_parameters,
    )


def _ncvar_to_dict(ncvar):
    """Convert a NetCDF Dataset variable to a dictionary."""
    # copy all attribute except for scaling parameters
    d = {
        k: getattr(ncvar, k)
        for k in ncvar.ncattrs()
        if k not in ["scale_factor", "add_offset"]
    }
    d["data"] = ncvar[:]
    if np.isscalar(d["data"]):
        # netCDF4 1.1.0+ returns a scalar for 0-dim array, we always want
        # 1-dim+ arrays with a valid shape.
        d["data"] = np.array(d["data"])
        d["data"].shape = (1,)
    return d



""" Unit Tests for Py-ART's io/cfradial.py module. """

import warnings

import netCDF4
import numpy as np
import pytest
from numpy.ma.core import MaskedArray
from numpy.testing import assert_almost_equal, assert_array_equal, assert_warns

import pyart

#################################################
# read_cfradial tests (verify radar attributes) #
#################################################

# read in the sample file and create a a Radar object
radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)


# time attribute
def test_time():
    assert "comment" in radar.time.keys()
    assert "long_name" in radar.time.keys()
    assert "standard_name" in radar.time.keys()
    assert "units" in radar.time.keys()
    assert "calendar" in radar.time.keys()
    assert "data" in radar.time.keys()
    assert radar.time["units"] == "seconds since 2011-05-20T10:54:08Z"
    assert radar.time["data"].shape == (40,)
    assert_almost_equal(radar.time["data"][38], 7, 0)


# range attribute
def test_range():
    assert "long_name" in radar.range
    assert "standard_name" in radar.range
    assert "meters_to_center_of_first_gate" in radar.range
    assert "meters_between_gates" in radar.range
    assert "units" in radar.range
    assert "data" in radar.range
    assert "spacing_is_constant" in radar.range
    assert radar.range["data"].shape == (42,)
    assert_almost_equal(radar.range["data"][0], 0, 0)


# fields attribute is tested later


# metadata attribute
def test_metadata():
    assert "Conventions" in radar.metadata
    assert "comment" in radar.metadata
    assert "history" in radar.metadata
    assert "institution" in radar.metadata
    assert "instrument_name" in radar.metadata
    assert "instrument_type" in radar.metadata
    assert "platform_type" in radar.metadata
    assert "primary_axis" in radar.metadata
    assert "references" in radar.metadata
    assert "source" in radar.metadata
    assert "title" in radar.metadata
    assert "version" in radar.metadata
    assert "volume_number" in radar.metadata


# scan_type attribute
def test_scan_type():
    assert radar.scan_type == "ppi"


# latitude attribute
def test_latitude():
    assert "data" in radar.latitude
    assert "standard_name" in radar.latitude
    assert "units" in radar.latitude
    assert radar.latitude["data"].shape == (1,)
    assert_almost_equal(radar.latitude["data"], 36.0, 0)


# longitude attribute
def test_longitude():
    assert "data" in radar.longitude
    assert "standard_name" in radar.longitude
    assert "units" in radar.longitude
    assert radar.longitude["data"].shape == (1,)
    assert_almost_equal(radar.longitude["data"], -98, 0)


# altitude attribute
def test_altitude():
    assert "data" in radar.altitude
    assert "standard_name" in radar.altitude
    assert "units" in radar.altitude
    assert "positive" in radar.altitude
    assert radar.altitude["data"].shape == (1,)
    assert_almost_equal(radar.altitude["data"], 214, 0)


# altitude_agl attribute
def test_altitude_agl():
    assert radar.altitude_agl is None


# sweep_number attribute
def test_sweep_number():
    assert "standard_name" in radar.sweep_number
    assert np.all(radar.sweep_number["data"] == range(1))


# sweep_mode attribute
def test_sweep_mode():
    assert "standard_name" in radar.sweep_mode
    assert radar.sweep_mode["data"].shape == (1, 32)
    assert radar.sweep_mode["data"].dtype.char == "S"
    str_array = netCDF4.chartostring(radar.sweep_mode["data"])
    try:
        assert np.all(str_array == ["azimuth_surveillance"])
    except AssertionError:
        assert np.all(str_array == [b"azimuth_surveillance"])


# fixed_angle attribute
def test_fixed_angle():
    assert "standard_name" in radar.fixed_angle
    assert "units" in radar.fixed_angle
    assert radar.fixed_angle["data"].shape == (1,)
    assert_almost_equal(radar.fixed_angle["data"][0], 0.50, 2)


# sweep_start_ray_index attribute
def test_sweep_start_ray_index():
    assert "long_name" in radar.sweep_start_ray_index
    assert radar.sweep_start_ray_index["data"].shape == (1,)
    assert_almost_equal(radar.sweep_start_ray_index["data"][0], 0, 0)


# sweep_end_ray_index attribute
def test_sweep_end_ray_index():
    assert "long_name" in radar.sweep_end_ray_index
    assert radar.sweep_end_ray_index["data"].shape == (1,)
    assert_almost_equal(radar.sweep_end_ray_index["data"][0], 39, 0)


# target_scan_rate attribute
def test_target_scan_rate():
    assert radar.target_scan_rate is None


# azimuth attribute
def test_azimuth():
    assert "standard_name" in radar.azimuth
    assert "units" in radar.azimuth
    assert_almost_equal(radar.azimuth["data"][0], 360, 0)
    assert_almost_equal(radar.azimuth["data"][10], 90.0, 0)


# elevation attribute
def test_elevation():
    assert "standard_name" in radar.elevation
    assert "units" in radar.elevation
    assert radar.elevation["data"].shape == (40,)
    assert_almost_equal(radar.elevation["data"][0], 0.48, 2)


# scan_rate attribute
def test_scan_rate():
    assert radar.scan_rate is None


# antenna_transition attribute
def test_antenna_transition():
    assert radar.antenna_transition is None


# instrument_parameters attribute
@pytest.mark.parametrize(
    "keys", ["prt_mode", "prt", "nyquist_velocity", "unambiguous_range"]
)
def test_instument_parameters(keys):
    # instrument_parameter sub-convention
    description = "instrument_parameters: %s" % keys
    check_instrument_parameter.description = description
    check_instrument_parameter(keys)


def check_instrument_parameter(param):
    assert param in radar.instrument_parameters
    param_dic = radar.instrument_parameters[param]
    assert param_dic["meta_group"] == "instrument_parameters"


# ngates attribute
def test_ngates():
    assert radar.ngates == 42


# nrays attribute
def test_nrays():
    assert radar.nrays == 40


# nsweeps attribute
def test_nsweeps():
    assert radar.nsweeps == 1


####################
# fields attribute #
####################


@pytest.mark.parametrize(
    "field",
    [
        "reflectivity_horizontal",
    ],
)
def test_field_dics(field):
    description = "field : %s, dictionary" % field
    check_field_dic.description = description
    check_field_dic(field)


def check_field_dic(field):
    """Check that the required keys are present in a field dictionary."""
    assert "standard_name" in radar.fields[field]
    assert "units" in radar.fields[field]
    assert "_FillValue" in radar.fields[field]
    assert "coordinates" in radar.fields[field]


@pytest.mark.parametrize(
    "field",
    [
        "reflectivity_horizontal",
    ],
)
def test_field_shapes(field):
    description = "field : %s, shape" % field
    check_field_shape.description = description
    check_field_shape(field)


def check_field_shape(field):
    assert radar.fields[field]["data"].shape == (40, 42)


fields = {
    "reflectivity_horizontal": MaskedArray,
}


@pytest.mark.parametrize("field, field_type", fields.items(), ids=list(fields.keys()))
def test_field_types(field, field_type):
    description = "field : %s, type" % field
    check_field_type.description = description
    check_field_type(field, field_type)


def check_field_type(field, field_type):
    assert type(radar.fields[field]["data"]) is field_type


fields = {
    "reflectivity_horizontal": -6.0,
}


@pytest.mark.parametrize("field, field_value", fields.items(), ids=list(fields.keys()))
def test_field_first_points(field, field_value):
    # these values can be found using:
    # [round(radar.fields[f]['data'][0,0]) for f in radar.fields]
    description = "field : %s, first point" % field
    check_field_first_point.description = description
    check_field_first_point(field, field_value)


def check_field_first_point(field, value):
    assert_almost_equal(radar.fields[field]["data"][0, 0], value, 0)


########################################################################
# write_cfradial tests (verify data in written netCDF matches original #
########################################################################


def test_write_ppi():
    # CF/Radial example file -> Radar object -> netCDF file
    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_ppi.nc"
        radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
        pyart.io.write_cfradial(tmpfile, radar)
        ref = netCDF4.Dataset(pyart.testing.CFRADIAL_PPI_FILE)
        dset = netCDF4.Dataset(tmpfile)
        check_dataset_to_ref(dset, ref)
        dset.close()


def test_write_ppi_fields_list():
    # Test if fields parameter is being used for writing specific fields.
    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_ppi.nc"
        tmpfile_warn = "tmp_ppi_warn.nc"
        include_fields = ["reflectivity_horizontal"]
        _format = "NETCDF4"
        radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
        pyart.io.write_cfradial(tmpfile, radar, include_fields=include_fields)
        ref = netCDF4.Dataset(pyart.testing.CFRADIAL_PPI_FILE)
        dset = netCDF4.Dataset(tmpfile)
        check_dataset_to_ref(dset, ref)
        include_fields = ["foo"]
        assert_warns(
            UserWarning,
            pyart.io.write_cfradial,
            tmpfile_warn,
            radar,
            _format,
            include_fields,
        )
        dset.close()


def test_write_ppi_unicode():
    # GitHub issue #381, unicode dtypes cause write to fail
    # CF/Radial example file -> Radar object -> netCDF file
    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_ppi_unicode.nc"
        radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
        # force the sweep mode array to be of unicode
        radar.sweep_mode["data"] = np.array(["azimuth_surveillance"], dtype="U")
        pyart.io.write_cfradial(tmpfile, radar)
        ref = netCDF4.Dataset(pyart.testing.CFRADIAL_PPI_FILE)
        dset = netCDF4.Dataset(tmpfile)
        check_dataset_to_ref(dset, ref)
        dset.close()


def test_write_ppi_U1():
    # GitHub issue #381, unicode dtypes cause write to fail
    # CF/Radial example file -> Radar object -> netCDF file
    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_ppi_unicode.nc"
        radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
        # force the sweep mode array to be a masked U1 array
        radar.sweep_mode["data"] = radar.sweep_mode["data"].astype("U")
        pyart.io.write_cfradial(tmpfile, radar)
        ref = netCDF4.Dataset(pyart.testing.CFRADIAL_PPI_FILE)
        dset = netCDF4.Dataset(tmpfile)
        check_dataset_to_ref(dset, ref)
        dset.close()


def test_write_ppi_unknown_instrument_parameter_element():
    # CF/Radial example file -> Radar object -> netCDF file
    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_ppi_unknonw_ip.nc"
        radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
        radar.instrument_parameters["foobar"] = {"data": np.zeros(40)}
        pytest.warns(UserWarning, pyart.io.write_cfradial, tmpfile, radar)
        ref = netCDF4.Dataset(pyart.testing.CFRADIAL_PPI_FILE)
        dset = netCDF4.Dataset(tmpfile)
        check_dataset_to_ref(dset, ref)
        dset.close()


def test_write_rhi():
    # CF/Radial example file -> Radar object -> netCDF file
    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_rhi.nc"
        radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_RHI_FILE)
        pyart.io.write_cfradial(tmpfile, radar)
        ref = netCDF4.Dataset(pyart.testing.CFRADIAL_RHI_FILE)
        dset = netCDF4.Dataset(tmpfile)
        check_dataset_to_ref(dset, ref)
        dset.close()


def test_write_ppi_arm_time_vars():
    # CF/Radial example file -> Radar object -> netCDF file
    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_arm_time_vars.nc"
        radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
        pyart.io.write_cfradial(tmpfile, radar, arm_time_variables=True)
        dset = netCDF4.Dataset(tmpfile)
        assert "base_time" in dset.variables
        assert "time_offset" in dset.variables

        base_time = dset.variables["base_time"]
        assert base_time[:] == 1305888856
        # The month is abbreviated according to the current locale, for
        # en_US the complete string is 20-May-2011,10:54:16 GMT.
        # In order to support other locales test only the non-locale dependent
        # portions of the string.
        assert base_time.string.startswith("20-")
        assert base_time.string.endswith("2011,10:54:16 GMT")

        time_offset = dset.variables["time_offset"]
        assert time_offset.units == "seconds since 2011-05-20 10:54:08"
        assert_almost_equal(time_offset[10], 11, 0)
        dset.close()


def check_dataset_to_ref(dset, ref):
    """Check that all data in Dataset is contained in the ref Dataset."""

    # file format (these are not expected to match)
    assert dset.file_format == ref.file_format

    # global attributes
    dset_attributes = dset.ncattrs()
    ref_attributes = ref.ncattrs()

    for attr in dset_attributes:
        print("Global attribute:", attr)
        assert attr in ref_attributes
        attribute_equal(dset, ref, attr, allow_str_case_diff=True)

    # cmptypes, expected to be empty
    print("Checking cmptypes...")
    assert dset.cmptypes == ref.cmptypes

    # groups, expected to be empty
    print("Checking groups...")
    assert dset.groups == ref.groups

    # dimensions (these are not expected to match)
    print("Checking dimensions")
    for dim in dset.dimensions.keys():
        print("Dimension", dim)
        assert dim in ref.dimensions.keys()

    # variables
    print("Checking variables...")
    dset_vars = dset.variables
    ref_vars = ref.variables
    for v in dset_vars.keys():
        if v in ["platform_type", "instrument_type", "primary_axis"]:
            continue  # default value created by pyart.
        print("Variable", v)
        check_variable_to_ref(dset_vars[v], ref_vars[v])


def check_variable_to_ref(var, ref_var):
    """Check that the data/metadata in var matches the ref. variable."""
    # check variable attributes
    for attr in var.ncattrs():
        print("Checking attribute", attr)
        assert attr in ref_var.ncattrs()
        attribute_equal(var, ref_var, attr)

    # instance variables
    assert var.dimensions == ref_var.dimensions
    assert var.dtype == ref_var.dtype
    assert var.ndim == ref_var.ndim
    assert var.shape == ref_var.shape

    if "least_significant_digit" in dir(var):
        assert "least_significant_digit" in dir(ref_var)
        assert var.least_significant_digit == ref_var.least_significant_digit

    # properties
    assert var.size == ref_var.size
    # netCDF4 version < 1.1.1 use the maskandscale attribute
    if hasattr(ref_var, "maskandscale"):
        assert var.maskandscale == ref_var.maskandscale
    # netCDF4 version >= 1.1.1 use a seperate mask and scale attributes
    if hasattr(ref_var, "mask"):
        assert var.mask == ref_var.mask
    if hasattr(ref_var, "scale"):
        assert var.scale == ref_var.scale

    # data and the mask
    data = var[:]
    ref_data = ref_var[:]
    assert ("mask" in dir(data)) == ("mask" in dir(ref_data))
    if "mask" in dir(data):
        assert_array_equal(data, ref_data)
        assert_array_equal(data.mask, ref_data.mask)
    else:
        assert_array_equal(data, ref_data)


def attribute_equal(class1, class2, key, allow_str_case_diff=True):
    """Check that an attribute of two classes are equal."""
    a1 = getattr(class1, key)
    a2 = getattr(class2, key)

    assert type(a1) == type(a2)

    if type(a1) is str and allow_str_case_diff:
        assert a1.upper() == a2.upper()
    else:
        assert a1 == a2
    return


def test_auto_history_and_conventions():
    # history and Conventions metadata should be created on write if
    # they do not exist in the original radar object
    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_auto_history.nc"
        radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_RHI_FILE)
        radar.metadata.pop("Conventions")
        radar.metadata.pop("history")
        pyart.io.write_cfradial(tmpfile, radar)
        radar2 = pyart.io.read_cfradial(tmpfile)
        assert "Conventions" in radar2.metadata
        assert "history" in radar2.metadata


def test_rhi_sweep_end_ray_index():
    radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_RHI_FILE)
    assert "long_name" in radar.sweep_end_ray_index
    assert radar.sweep_end_ray_index["data"].shape == (1,)
    assert_almost_equal(radar.sweep_end_ray_index["data"][0], 39, 0)


def test_delay_field_loading():
    radar = pyart.io.read_cfradial(
        pyart.testing.CFRADIAL_PPI_FILE, delay_field_loading=True
    )
    assert isinstance(
        radar.fields["reflectivity_horizontal"], pyart.lazydict.LazyLoadDict
    )
    data = radar.fields["reflectivity_horizontal"]["data"]
    assert isinstance(data, MaskedArray)
    assert data.shape == (40, 42)
    assert_almost_equal(data[0, 0], -6.0, 0)


def test_create_ncvar_different_dtype():
    # test _Write_as_dtype key handling in _create_ncvar
    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_nvcar_different_dtype.nc"
        dset = netCDF4.Dataset(tmpfile, mode="w")
        dset.createDimension("x", 256)
        dic = {
            "data": np.arange(256, dtype="float32") * 0.5 + 100.0,
            "_Write_as_dtype": "u1",
            "_FillValue": 100.0,
        }
        assert "add_offset" not in dic
        assert "scale_factor" not in dic

        pyart.io.cfradial._create_ncvar(dic, dset, "foo", ("x"))

        assert "add_offset" in dic
        assert "scale_factor" in dic
        assert "_FillValue" in dic

        foo = dset.variables["foo"]
        assert foo.dtype == np.dtype("uint8")
        assert_almost_equal(foo.scale_factor, 0.5)
        assert_almost_equal(foo.add_offset, 100)
        assert foo[0] is np.ma.masked
        assert_almost_equal(foo[1], 100.5)
        assert_almost_equal(foo[-1], 227.5)

        dset.close()


def test_calculate_scale_and_offset():
    data = np.arange(128, dtype="float32")
    scale, offset, fill = pyart.io.cfradial._calculate_scale_and_offset(
        {"data": data}, np.dtype("u1")
    )

    assert_almost_equal(scale, 0.5)
    assert_almost_equal(offset, -0.5)
    assert_almost_equal(fill, 0)

    # check that nans and infs are ignored
    data[5] = np.nan
    data[6] = np.inf
    scale, offset, fill = pyart.io.cfradial._calculate_scale_and_offset(
        {"data": data}, np.dtype("u1")
    )
    assert_almost_equal(scale, 0.5)
    assert_almost_equal(offset, -0.5)
    assert_almost_equal(fill, 0)

    # maximum < minimum raises ValueError
    pytest.raises(
        ValueError,
        pyart.io.cfradial._calculate_scale_and_offset,
        {"data": data},
        np.dtype("u1"),
        100.0,
        99,
    )

    # maximum == minimum issues warning
    with warnings.catch_warnings(record=True) as w:
        assert len(w) == 0
        pyart.io.cfradial._calculate_scale_and_offset(
            {"data": data}, np.dtype("u1"), 100, 100
        )
        assert len(w) == 1



""" Unit tests for the xsect.py module. """

from numpy.testing import assert_almost_equal

import pyart


def test_cross_section_ppi():
    radar = pyart.testing.make_target_radar()
    xsect = pyart.util.cross_section_ppi(radar, [45, 60])

    assert xsect.nsweeps == 2
    assert xsect.nrays == 2
    assert len(xsect.time["data"]) == 2
    assert_almost_equal(xsect.azimuth["data"][0], 45)
    assert_almost_equal(xsect.azimuth["data"][1], 60)
    assert_almost_equal(xsect.elevation["data"][0], 0.75, 2)
    assert_almost_equal(xsect.elevation["data"][1], 0.75, 2)
    assert xsect.fields["reflectivity"]["data"].shape == (2, 50)
    assert len(xsect.sweep_number["data"]) == 2
    assert len(xsect.sweep_mode["data"]) == 2
    assert len(xsect.fixed_angle["data"]) == 2
    assert xsect.scan_type == "rhi"
    assert_almost_equal(xsect.sweep_start_ray_index["data"][0], 0)
    assert_almost_equal(xsect.sweep_start_ray_index["data"][1], 1)
    assert_almost_equal(xsect.sweep_end_ray_index["data"][0], 0)
    assert_almost_equal(xsect.sweep_end_ray_index["data"][1], 1)



#! /usr/bin/env python
"""
Make a small sigmet (IRIS) file containing a single RHI scan.

Reflectivity data is set to 0 dBZ, other parameters are taken from the
full sized file XSW110520113537.RAW7HHL
"""

import struct

import numpy as np

# parameters
INFILE = "XSW110520113537.RAW7HHL"
OUTFILE = "example_sigmet_rhi.sigmet"

# open the input and output files
f = open(INFILE, "rb")
out = open(OUTFILE, "wb")

#############################
# First Record (6144 bytes) #
# product_hdr (640 bytes)   #
#############################

# structure_header (12 bytes)
out.write(f.read(12))  # no change

# product_configuration (320 bytes)
# -> structure_header
# -> generation_time
# -> sweep_ingest_time
# -> file_ingest_time
# -> color_scale_def
out.write(f.read(320))  # no change

# product_end (308 bytes)
# -> ingest_time
fmt = (
    "16s8s8s12s28sh16s16shIIhhiiHHh12sHiiiiiHhHhhhHhhhhHH18sIIIIIIHHIIh"
    + "32shhhBB2s8sI4s"
)
packet = list(struct.unpack(fmt, f.read(308)))
packet[23] = 154000  # last_bin_range
packet[24] = 25  # number_bins
out.write(struct.pack(fmt, *packet))

# padding to next record
out.write(f.read(6144 - 640))  # padding to next record

##############################
# Second Record (6144 bytes) #
# ingest_header (4884 bytes) #
##############################
# structure_header
out.write(f.read(12))  # no change

# ingest_configuration (480 bytes)
# -> volume_scan_start_time
# ->
fmt = "80shhi12s12shhhh4s8s16sh16shIIhhHHHhiiiiiiiIh2s8sI16s228s"
packet = list(struct.unpack(fmt, f.read(480)))
packet[2] = 1  # number_sweeps_completed
packet[22] = 20  # number_rays_sweep
out.write(struct.pack(fmt, *packet))

# task_configuration (2612 bytes)
# -> structure_header (12 byes)
out.write(f.read(12))  # no change

# -> task_sched_info (120 bytes)
out.write(f.read(120))  # no change

# -------------------------------------------------
# -> task_dsp_info (320 bytes)
# ---> major_mode, dsp_type (4 bytes)
out.write(f.read(4))

# ---> current_data_type_mask (24 bytes)
fmt = "IIIIII"
packet = list(struct.unpack(fmt, f.read(24)))
packet[0] = 512  # mask_word_0 indicating field 9 (DBZ2), 2 ** 9
packet[2] = 0  # mask_word_1 indicating no fields
out.write(struct.pack(fmt, *packet))

# ---> original_data_type_mask, task_dsp_mode, etc (292 bytes)
out.write(f.read(292))
# ---------------------------------------------------

# -> task_calib_info (320 bytes)
out.write(f.read(320))  # no change

# -> task_range_info (160 bytes) - file at 7408 seek
fmt = "iihhiiHh136s"
packet = list(struct.unpack(fmt, f.read(160)))
packet[1] = 154000  # last_bin_range
packet[2] = 25  # number_input_bins
packet[3] = 25  # number_output_bins
out.write(struct.pack(fmt, *packet))

# -> task_scan_info (320 bytes)
fmt = "Hh2sh200s112s"
packet = list(struct.unpack(fmt, f.read(320)))
packet[3] = 1  # number_sweeps
out.write(struct.pack(fmt, *packet))

# -> task_misc_info (320 bytes)
out.write(f.read(320))  # no change

# -> task_end_info (320 bytes)
out.write(f.read(320))  # no change

# -> comments (720 bytes)
out.write(f.read(720))  # no change


# spare_0 (732 bytes)
out.write(f.read(732))  # no change

# gparm (128 bytes)
out.write(f.read(128))  # no change

# reserved (920 bytes)
out.write(f.read(920))  # no change

# padding to next record
out.write(f.read(6144 - 4884))

###################
# Data            #
###################
# f is at 12288
# raw_prod_bhdr (12 bytes)
out.write(f.read(12))

# skip over ingest_data_header for field 8
f.read(76)

# ingest_data_header for field 9
fmt = "12s12shhhhhHhH36s"
packet = list(struct.unpack(fmt, f.read(76)))
packet[3] = 20  # number_rays_sweep
packet[5] = 20  # number_rays_file_expected
packet[6] = 20  # number_rays_file_actual
out.write(struct.pack(fmt, *packet))

# compressed ray data
code = np.ones((33), dtype="int16")
code[0] = -32737  # indicates 31 values follow (-32737 + 32768 = 31)
code[1] = 6790  # az0
code[3] = 6790  # az1
code[5] = 25  # nbins
code[6:-1] = -32768  # 0 dBZ
code[-1] = 1  # end of ray marker

for i in range(20):
    code[6] = i  # time
    code.view("uint16")[2] = 91 * 18 * i  # el0 0 -> 180 in steps of 18
    code.view("uint16")[4] = 91 * 18 * i  # el1
    out.write(code.tostring())

out.write("\x00" * 4736)

# close the files
f.close()
out.close()



""" Unit Tests for Py-ART's retrieve/simple_moment_calculation.py module. """

import numpy as np
from numpy.testing import assert_allclose

import pyart

# Setup a test radar object for various retrievals
test_radar = pyart.testing.make_empty_ppi_radar(100, 360, 5)
test_radar.range["data"] = test_radar.range["data"] * 100.0
range_grid = (
    np.meshgrid(test_radar.range["data"], np.ma.ones(test_radar.time["data"].shape))[0]
    + 1.0
)
foo_field = {"data": np.zeros([360 * 5, 100]) + 20.0 * np.log10(range_grid / 1000.0)}
test_radar.add_field("reflectivity", foo_field)


def test_calculate_snr_from_reflectivity():
    """Test the calculate_snr_from_reflectivity function from pyart.retrieve"""
    snr = pyart.retrieve.calculate_snr_from_reflectivity(test_radar, toa=500)
    assert snr["data"].mean() < 1e-6


def test_compute_noisedBZ():
    """Test the compute_noisedBZ function from pyart.retrieve"""
    noise = pyart.retrieve.compute_noisedBZ(
        nrays=test_radar.nrays,
        noisedBZ_val=1.0,
        _range=test_radar.range["data"],
        ref_dist=1.0,
        noise_field="noisedBZ_hh",
    )
    test_radar.add_field_like
    assert noise["data"].min() > 1.0
    assert noise["data"].max() < 45.0


def test_compute_snr():
    noise = pyart.retrieve.compute_noisedBZ(
        nrays=test_radar.nrays,
        noisedBZ_val=1.0,
        _range=test_radar.range["data"],
        ref_dist=1.0,
        noise_field="noisedBZ_hh",
    )
    # Add the noise field to the radar
    test_radar.add_field("noisedBZ_hh", noise)

    # Calculate snr from reflectivity and noise
    snr = pyart.retrieve.compute_snr(
        test_radar, refl_field="reflectivity", noise_field="noisedBZ_hh"
    )
    assert snr["data"].min() > -1.0


def test_calculate_velocity_texture():
    radar = pyart.testing.make_empty_ppi_radar(10, 36, 1)

    # a zero field
    fdata = np.tile(np.arange(10.0), 36).reshape(36, 10)
    fdata3 = np.zeros(fdata.shape)
    radar.add_field("zero_field", {"data": fdata3})

    vel_field = "zero_field"
    texture_field = pyart.retrieve.calculate_velocity_texture(
        radar, vel_field, wind_size=2, nyq=10
    )
    assert np.all(texture_field["data"] == 0)
    texture_field = pyart.retrieve.calculate_velocity_texture(
        radar, vel_field, wind_size=3, nyq=10
    )
    assert np.all(texture_field["data"] == 0)
    texture_field = pyart.retrieve.calculate_velocity_texture(
        radar, vel_field, wind_size=4, nyq=10
    )
    assert np.all(texture_field["data"] == 0)

    # Test none parameters
    radar2 = pyart.io.read(pyart.testing.NEXRAD_ARCHIVE_MSG31_FILE)
    texture_field = pyart.retrieve.calculate_velocity_texture(
        radar2, vel_field=None, wind_size=4, nyq=None
    )
    assert_allclose(
        texture_field["data"][-1][-5:],
        [0.000363, 0.000363, 0.000363, 0.000363, 0.000363],
        atol=1e-03,
    )



"""
Corrects polarimetric variables for noise

"""

import numpy as np

from ..config import get_field_name, get_metadata


def correct_noise_rhohv(
    radar,
    urhohv_field=None,
    snr_field=None,
    zdr_field=None,
    nh_field=None,
    nv_field=None,
    rhohv_field=None,
):
    """
    Corrects RhoHV for noise according to eq. 6 in Gourley et al. 2006.
    This correction should only be performed if noise has not been subtracted
    from the signal during the moments computation.

    Parameters
    ----------
    radar : Radar
        Radar object.
    urhohv_field : str, optional
        Name of the RhoHV uncorrected for noise field.
    snr_field, zdr_field, nh_field, nv_field : str, optional
        Names of the SNR, ZDR, horizontal channel noise in dBZ and vertical
        channel noise in dBZ used to correct RhoHV.
    rhohv_field : str, optional
        Name of the rhohv field to output.

    Returns
    -------
    rhohv : dict
        Noise corrected RhoHV field.

    References
    ----------
    Gourley et al. Data Quality of the Meteo-France C-Band Polarimetric
    Radar, JAOT, 23, 1340-1356

    """
    # parse the field parameters
    if urhohv_field is None:
        urhohv_field = get_field_name("uncorrected_cross_correlation_ratio")
    if snr_field is None:
        snr_field = get_field_name("signal_to_noise_ratio")
    if zdr_field is None:
        zdr_field = get_field_name("differential_reflectivity")
    if nh_field is None:
        nh_field = get_field_name("noisedBZ_hh")
    if nv_field is None:
        nv_field = get_field_name("noisedBZ_vv")
    if rhohv_field is None:
        rhohv_field = get_field_name("cross_correlation_ratio")

    # extract fields from radar
    if urhohv_field in radar.fields:
        urhohv = radar.fields[urhohv_field]["data"]
    else:
        raise KeyError("Field not available: " + urhohv_field)
    if snr_field in radar.fields:
        snrdB_h = radar.fields[snr_field]["data"]
    else:
        raise KeyError("Field not available: " + snr_field)
    if zdr_field in radar.fields:
        zdrdB = radar.fields[zdr_field]["data"]
    else:
        raise KeyError("Field not available: " + zdr_field)
    if nh_field in radar.fields:
        nh = radar.fields[nh_field]["data"]
    else:
        raise KeyError("Field not available: " + nh_field)
    if nv_field in radar.fields:
        nv = radar.fields[nv_field]["data"]
    else:
        raise KeyError("Field not available: " + nv_field)

    snr_h = np.ma.power(10.0, 0.1 * snrdB_h)
    zdr = np.ma.power(10.0, 0.1 * zdrdB)
    alpha = np.ma.power(10.0, 0.1 * (nh - nv))

    rhohv_data = urhohv * np.ma.sqrt(
        (1.0 + 1.0 / snr_h) * (1.0 + zdr / (alpha * snr_h))
    )
    rhohv_data[rhohv_data > 1.0] = 1.0

    rhohv = get_metadata(rhohv_field)
    rhohv["data"] = rhohv_data

    return rhohv


def correct_bias(radar, bias=0.0, field_name=None):
    """
    Corrects a radar data bias. If field name is none the correction is
    applied to horizontal reflectivity by default.

    Parameters
    ----------
    radar : Radar
        Radar object.
    bias : float, optional
        The bias magnitude.
    field_name: str, optional
        Names of the field to be corrected.

    Returns
    -------
    corrected_field : dict
        The corrected field

    """
    # parse the field parameters
    if field_name is None:
        field_name = get_field_name("reflectivity")

    # extract fields from radar
    if field_name in radar.fields:
        field_data = radar.fields[field_name]["data"]
    else:
        raise KeyError("Field not available: " + field_name)

    corr_field_data = field_data - bias

    if field_name.startswith("corrected_"):
        corr_field_name = field_name
    else:
        corr_field_name = "corrected_" + field_name

    corr_field = get_metadata(corr_field_name)
    corr_field["data"] = corr_field_data

    return corr_field



"""
Routines for reading sinarame_H5 files.

"""

import glob
import os
from datetime import datetime

try:
    from netcdftime import num2date
except ImportError:
    from cftime import num2date

import numpy as np

try:
    import h5py

    _H5PY_AVAILABLE = True
except ImportError:
    _H5PY_AVAILABLE = False

from ..config import FileMetadata, get_fillvalue
from ..core.radar import Radar
from ..exceptions import MissingOptionalDependency
from ..io import write_cfradial
from ..io.common import _test_arguments, make_time_unit_str

SINARAME_H5_FIELD_NAMES = {
    "TH": "total_power",  # uncorrected reflectivity, horizontal
    "TV": "total_power_vertical",  # uncorrected reflectivity, vertical
    "DBZH": "reflectivity",  # corrected reflectivity, horizontal
    "DBZV": "reflectivity",  # corrected reflectivity, vertical
    "CM": "clutter mask",
    "ZDR": "differential_reflectivity",  # differential reflectivity
    "RHOHV": "cross_correlation_ratio",
    "LDR": "linear_polarization_ratio",
    "PHIDP": "differential_phase",
    "KDP": "specific_differential_phase",
    "SQI": "normalized_coherent_power",
    "SNR": "signal_to_noise_ratio",
    "VRAD": "velocity",
    "WRAD": "spectrum_width",
    "QIND": "quality_index",
}


def read_sinarame_h5(
    filename,
    field_names=None,
    additional_metadata=None,
    file_field_names=False,
    exclude_fields=None,
    include_fields=None,
    **kwargs,
):
    """
    Read a SINARAME_H5 file.

    Parameters
    ----------
    filename : str
        Name of the SINARAME_H5 file to read.
    field_names : dict, optional
        Dictionary mapping SINARAME_H5 field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.


    Returns
    -------
    radar : Radar
        Radar object containing data from SINARAME_H5 file.

    """
    # TODO before moving to pyart.io
    # * unit test
    # * add default field mapping, etc to default config
    # * auto-detect file type with pyart.io.read function
    # * instrument parameters
    # * add additional checks for HOW attributes
    # * support for other objects (SCAN, XSEC)

    # check that h5py is available
    if not _H5PY_AVAILABLE:
        raise MissingOptionalDependency(
            "h5py is required to use read_sinarame_h5 but is not installed"
        )

    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    if field_names is None:
        field_names = SINARAME_H5_FIELD_NAMES
    filemetadata = FileMetadata(
        "SINARAME_h5",
        field_names,
        additional_metadata,
        file_field_names,
        exclude_fields,
        include_fields,
    )

    # open the file
    hfile = h5py.File(filename, "r")
    SINARAME_object = _to_str(hfile["what"].attrs["object"])
    if SINARAME_object not in ["PVOL", "SCAN", "ELEV", "AZIM"]:
        raise NotImplementedError("object: %s not implemented." % (SINARAME_object))

    # determine the number of sweeps by the number of groups which
    # begin with dataset
    datasets = [k for k in hfile if k.startswith("dataset")]
    datasets.sort(key=lambda x: int(x[7:]))
    nsweeps = len(datasets)

    # latitude, longitude and altitude
    latitude = filemetadata("latitude")
    longitude = filemetadata("longitude")
    altitude = filemetadata("altitude")

    h_where = hfile["where"].attrs
    latitude["data"] = np.array([h_where["lat"]], dtype="float64")
    longitude["data"] = np.array([h_where["lon"]], dtype="float64")
    altitude["data"] = np.array([h_where["height"]], dtype="float64")

    # metadata
    metadata = filemetadata("metadata")
    metadata["source"] = _to_str(hfile["what"].attrs["source"])
    metadata["original_container"] = "SINARAME_h5"
    metadata["SINARAME_conventions"] = _to_str(hfile.attrs["Conventions"])

    h_what = hfile["what"].attrs
    metadata["version"] = _to_str(h_what["version"])
    metadata["source"] = _to_str(h_what["source"])

    try:
        ds1_how = hfile[datasets[0]]["how"].attrs
    except KeyError:
        # if no how group exists mock it with an empty dictionary
        ds1_how = {}
    if "system" in ds1_how:
        metadata["system"] = ds1_how["system"]
    if "software" in ds1_how:
        metadata["software"] = ds1_how["software"]
    if "sw_version" in ds1_how:
        metadata["sw_version"] = ds1_how["sw_version"]

    # sweep_start_ray_index, sweep_end_ray_index
    sweep_start_ray_index = filemetadata("sweep_start_ray_index")
    sweep_end_ray_index = filemetadata("sweep_end_ray_index")

    if SINARAME_object in ["AZIM", "SCAN", "PVOL"]:
        rays_per_sweep = [int(hfile[d]["where"].attrs["nrays"]) for d in datasets]
    elif SINARAME_object == "ELEV":
        rays_per_sweep = [int(hfile[d]["where"].attrs["angles"].size) for d in datasets]
    total_rays = sum(rays_per_sweep)
    ssri = np.cumsum(np.append([0], rays_per_sweep[:-1])).astype("int32")
    seri = np.cumsum(rays_per_sweep).astype("int32") - 1
    sweep_start_ray_index["data"] = ssri
    sweep_end_ray_index["data"] = seri

    # sweep_number
    sweep_number = filemetadata("sweep_number")
    sweep_number["data"] = np.arange(nsweeps, dtype="int32")

    # sweep_mode
    sweep_mode = filemetadata("sweep_mode")
    sweep_mode["data"] = np.array(nsweeps * ["azimuth_surveillance"])

    # scan_type
    if SINARAME_object == "ELEV":
        scan_type = "rhi"
    else:
        scan_type = "ppi"

    # fixed_angle
    fixed_angle = filemetadata("fixed_angle")
    if SINARAME_object == "ELEV":
        sweep_el = [hfile[d]["where"].attrs["az_angle"] for d in datasets]
    else:
        sweep_el = [hfile[d]["where"].attrs["elangle"] for d in datasets]
    fixed_angle["data"] = np.array(sweep_el, dtype="float32")

    # elevation
    elevation = filemetadata("elevation")
    if "elangles" in ds1_how:
        edata = np.empty(total_rays, dtype="float32")
        for d, start, stop in zip(datasets, ssri, seri):
            edata[start : stop + 1] = hfile[d]["how"].attrs["elangles"][:]
        elevation["data"] = edata
    elif SINARAME_object == "ELEV":
        edata = np.empty(total_rays, dtype="float32")
        for d, start, stop in zip(datasets, ssri, seri):
            edata[start : stop + 1] = hfile[d]["where"].attrs["angles"][:]
        elevation["data"] = edata
    else:
        elevation["data"] = np.repeat(sweep_el, rays_per_sweep)

    # range
    _range = filemetadata("range")
    if "rstart" in hfile["dataset1/where"].attrs:
        # derive range from rstart and rscale attributes if available

        # check that the gate spacing is constant between sweeps
        rstart = [hfile[d]["where"].attrs["rstart"] for d in datasets]
        if any(rstart != rstart[0]):
            raise ValueError("range start changes between sweeps")
        rscale = [hfile[d]["where"].attrs["rscale"] for d in datasets]
        if any(rscale != rscale[0]):
            raise ValueError("range scale changes between sweeps")
        nbins = int(hfile["dataset1"]["where"].attrs["nbins"])
        _range["data"] = np.arange(nbins, dtype="float32") * rscale[0] + rstart[0]
        _range["meters_to_center_of_first_gate"] = rstart[0]
        _range["meters_between_gates"] = float(rscale[0])
    else:
        # if not defined use range attribute which defines the maximum range in
        # km.  There is no information on the starting location of the range
        # bins so we assume this to be 0.
        # This most often occurs in RHI files, which technically do not meet
        # the ODIM 2.2 specs. Section 7.4 requires that these files include
        # the where/rstart, where/rscale and where/nbins attributes.
        max_range = [hfile[d]["where"].attrs["range"] for d in datasets]
        if any(max_range != max_range[0]):
            raise ValueError("maximum range changes between sweeps")
        # nbins is required
        nbins = hfile["dataset1/data1/data"].shape[1]
        _range["data"] = np.linspace(0, max_range[0] * 1000.0, nbins, dtype="float32")
        _range["meters_to_center_of_first_gate"] = 0
        _range["meters_between_gates"] = max_range[0] * 1000.0 / nbins

    # azimuth
    azimuth = filemetadata("azimuth")
    az_data = np.ones((total_rays,), dtype="float32")
    for dset, start, stop in zip(datasets, ssri, seri):
        if SINARAME_object == "ELEV":
            # all azimuth angles are the sweep azimuth angle
            sweep_az = hfile[dset]["where"].attrs["az_angle"]
        elif SINARAME_object == "AZIM":
            # Sector azimuths are specified in the startaz and stopaz
            # attribute of dataset/where.
            # Assume that the azimuth angles do not pass through 0/360 degrees
            startaz = hfile[dset]["where"].attrs["startaz"]
            stopaz = hfile[dset]["where"].attrs["stopaz"]
            nrays = stop - start + 1
            sweep_az = np.linspace(startaz, stopaz, nrays, endpoint=True)
        elif ("startazA" in ds1_how) and ("stopazA" in ds1_how):
            # average between start and stop azimuth angles
            startaz = hfile[dset]["how"].attrs["startazA"]
            stopaz = hfile[dset]["how"].attrs["stopazA"]
            sweep_az = np.angle(
                (np.exp(1.0j * np.deg2rad(startaz)) + np.exp(1.0j * np.deg2rad(stopaz)))
                / 2.0,
                deg=True,
            )
        else:
            # according to section 5.1 the first ray points north (0 degrees)
            # and proceeds clockwise for a complete 360 rotation.
            nrays = stop - start + 1
            sweep_az = np.linspace(0, 360, nrays, endpoint=False)
        az_data[start : stop + 1] = sweep_az
    azimuth["data"] = az_data

    # time
    _time = filemetadata("time")
    if ("startazT" in ds1_how) and ("stopazT" in ds1_how):
        # average between startazT and stopazT
        t_data = np.empty((total_rays,), dtype="float32")
        for dset, start, stop in zip(datasets, ssri, seri):
            t_start = hfile[dset]["how"].attrs["startazT"]
            t_stop = hfile[dset]["how"].attrs["stopazT"]
            t_data[start : stop + 1] = (t_start + t_stop) / 2
        start_epoch = t_data.min()
        start_time = datetime.utcfromtimestamp(start_epoch)
        _time["units"] = make_time_unit_str(start_time)
        _time["data"] = t_data - start_epoch
    else:
        t_data = np.empty((total_rays,), dtype="int32")
        # interpolate between each sweep starting and ending time
        for dset, start, stop in zip(datasets, ssri, seri):
            dset_what = hfile[dset]["what"].attrs
            start_str = _to_str(dset_what["startdate"] + dset_what["starttime"])
            end_str = _to_str(dset_what["enddate"] + dset_what["endtime"])
            start_dt = datetime.strptime(start_str, "%Y%m%d%H%M%S")
            end_dt = datetime.strptime(end_str, "%Y%m%d%H%M%S")

            time_delta = end_dt - start_dt
            delta_seconds = time_delta.seconds + time_delta.days * 3600 * 24
            rays = stop - start + 1
            sweep_start_epoch = (start_dt - datetime(1970, 1, 1)).total_seconds()
            t_data[start : stop + 1] = sweep_start_epoch + np.linspace(
                0, delta_seconds, rays
            )
        start_epoch = t_data.min()
        start_time = datetime.utcfromtimestamp(start_epoch)
        _time["units"] = make_time_unit_str(start_time)
        _time["data"] = (t_data - start_epoch).astype("float32")

    # fields
    fields = {}
    h_field_keys = [k for k in hfile["dataset1"] if k.startswith("data")]
    SINARAME_fields = [
        hfile["dataset1"][d]["what"].attrs["quantity"] for d in h_field_keys
    ]
    for SINARAME_field, h_field_key in zip(SINARAME_fields, h_field_keys):
        field_name = filemetadata.get_field_name(_to_str(SINARAME_field))
        if field_name is None:
            continue
        fdata = np.ma.zeros((total_rays, nbins), dtype="float32")
        start = 0
        # loop over the sweeps, copy data into correct location in data array
        for dset, rays_in_sweep in zip(datasets, rays_per_sweep):
            sweep_data = _get_SINARAME_h5_sweep_data(hfile[dset][h_field_key])
            sweep_nbins = sweep_data.shape[1]
            fdata[start : start + rays_in_sweep, :sweep_nbins] = sweep_data[:]
            start += rays_in_sweep
        # create field dictionary
        field_dic = filemetadata(field_name)
        field_dic["data"] = fdata
        field_dic["_FillValue"] = get_fillvalue()
        fields[field_name] = field_dic
        # Add missing metadata
        if file_field_names:
            fields[field_name].update(
                filemetadata.get_metadata(field_names[field_name])
            )

    # instrument_parameters
    instrument_parameters = None

    return Radar(
        _time,
        _range,
        fields,
        metadata,
        scan_type,
        latitude,
        longitude,
        altitude,
        sweep_number,
        sweep_mode,
        fixed_angle,
        sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth,
        elevation,
        instrument_parameters=instrument_parameters,
    )


def write_sinarame_cfradial(path):
    """
    This function takes SINARAME_H5 files (where every file has only one field
    and one volume) from a folder and writes a CfRadial file for each volume
    including all fields.

    Parameters
    ----------
    path : str
        Where the SINARAME_H5 files are.

    """

    path_user = os.path.expanduser(path)

    TH_list = glob.glob(path_user + "/*_TH_*.H5")

    file_date = [i.split("_")[-1][9:-4] for i in TH_list]
    file_date.sort()

    for i in file_date:
        files = glob.glob(path_user + "/*" + i + "Z.H5")

        # I want it to start with TV or TH because of some range issue
        files.sort(reverse=True)
        files.sort(key=lambda x: len(x.split("_")[-2]))

        for j in np.arange(len(files)):
            basename = os.path.basename(files[j])
            bs = basename.split("_")
            base1 = "{b1}_{b2}_{b3}_{fn}_{b4}".format(
                b1=bs[0], b2=bs[1], b3=bs[2], fn=bs[3], b4=bs[4]
            )
            file = f"{path_user}/{base1}"

            if j == 0:
                try:
                    radar = read_sinarame_h5(file, file_field_names=True)
                    azi_shape, range_shape = radar.fields["TV"]["data"].shape

                except ValueError:
                    print("x - this Radar wasn't created", base1, sep="\t")

                # In case the H5 file was badly created with bbufr it
                # throws a KeyError
                except KeyError:
                    print("x - Wrong BUFR conversion", base1, sep="\t")

            else:
                try:
                    radar_tmp = read_sinarame_h5(file, file_field_names=True)

                    field = radar_tmp.fields.keys()[0]

                    # Add missing gates
                    if radar_tmp.fields[field]["data"].shape[1] != range_shape:
                        n_missing_gates = (
                            range_shape - radar_tmp.fields[field]["data"].shape[1]
                        )
                        fill = np.ma.masked_all((azi_shape, n_missing_gates))
                        data = np.ma.concatenate(
                            [radar_tmp.fields[field]["data"], fill], 1
                        )
                        radar_tmp.fields[field]["data"] = data

                    radar.fields.update(radar_tmp.fields)

                # Same as above, bbufr conversion errors
                except KeyError:
                    print("x - Wrong BUFR conversion", base1, sep="\t")

                except ValueError:
                    print("x - this Radar wasn't created", base1, sep="\t")

                # In case the first file didn't create a Radar object
                except NameError:
                    print("Radar didn't exist, creating")
                    radar = read_sinarame_h5(file, file_field_names=True)

        time1 = num2date(
            radar.time["data"][0],
            radar.time["units"],
            calendar="standard",
            only_use_cftime_datetimes=True,
            only_use_python_datetimes=False,
        ).strftime("%Y%m%d_%H%M%S")
        time2 = num2date(
            radar.time["data"][-1],
            radar.time["units"],
            calendar="standard",
            only_use_cftime_datetimes=True,
            only_use_python_datetimes=False,
        ).strftime("%Y%m%d_%H%M%S")

        radar._DeflateLevel = 5

        # cfrad.TIME1_to_TIME2_NAME_VCP_RANGE.nc
        cffile = "cfrad.{time1}.0000_to_{time2}.0000" "_{b1}_{est}_{ran}".format(
            time1=time1, time2=time2, b1=bs[0], est=bs[1], ran=bs[2]
        )

        print(f"Writing to {path_user}{cffile}.nc")
        write_cfradial(
            path_user + "/" + cffile + ".nc", radar, format="NETCDF4_CLASSIC"
        )


def _to_str(text):
    """Convert bytes to str if necessary."""
    if hasattr(text, "decode"):
        return text.decode("utf-8")
    else:
        return text


def _get_SINARAME_h5_sweep_data(group):
    """Get SINARAME_H5 sweet data from an HDF5 group."""

    # mask raw data
    what = group["what"]
    raw_data = group["data"][:]

    if "nodata" in what.attrs:
        nodata = what.attrs.get("nodata")
        data = np.ma.masked_equal(raw_data, nodata)
    else:
        data = np.ma.masked_array(raw_data)
    if "undetect" in what.attrs:
        undetect = what.attrs.get("undetect")
        data[data == undetect] = np.ma.masked

    offset = 0.0
    gain = 1.0
    if "offset" in what.attrs:
        offset = what.attrs.get("offset")
    if "gain" in what.attrs:
        gain = what.attrs.get("gain")
    return data * gain + offset



#! /usr/bin/env python
"""
Make a small netCDF CF/Radial file containing a single PPI scan.

Single field and scan is converted from sigmet file XSW110520105408.RAW7HHF
"""

import pyart

radar = pyart.io.read_rsl("XSW110520105408.RAW7HHF")

time_slice = slice(None, 400, 10)
range_slice = slice(None, None, 16)
sweep_slice = slice(None, 1)

# remove all but the reflectivity_horizontal fields
rf_field = radar.fields["reflectivity"]
rf_data = rf_field["data"]
rf_field["data"] = rf_data[time_slice, range_slice]
radar.fields = {"reflectivity_horizontal": rf_field}

radar.nsweeps = 1
radar.nray = 40
radar.ngates = 42

# truncate the range based variables
radar.range["data"] = radar.range["data"][range_slice]

# truncate the time based variables
radar.time["data"] = radar.time["data"][time_slice]
radar.azimuth["data"] = radar.azimuth["data"][time_slice]
radar.elevation["data"] = radar.elevation["data"][time_slice]
radar.instrument_parameters["prt"]["data"] = radar.instrument_parameters["prt"]["data"][
    time_slice
]

radar.instrument_parameters["unambiguous_range"]["data"] = radar.instrument_parameters[
    "unambiguous_range"
]["data"][time_slice]

radar.instrument_parameters["nyquist_velocity"]["data"] = radar.instrument_parameters[
    "nyquist_velocity"
]["data"][time_slice]

# truncate the sweep based variables
radar.sweep_number["data"] = radar.sweep_number["data"][sweep_slice]
radar.fixed_angle["data"] = radar.fixed_angle["data"][sweep_slice]
radar.sweep_start_ray_index["data"] = radar.sweep_start_ray_index["data"][sweep_slice]
radar.sweep_end_ray_index["data"] = radar.sweep_end_ray_index["data"][sweep_slice]
radar.sweep_end_ray_index["data"][0] = 39
radar.sweep_mode["data"] = radar.sweep_mode["data"][sweep_slice]

radar.sweep_number["data"] = radar.sweep_number["data"][sweep_slice]

radar.instrument_parameters["prt_mode"]["data"] = radar.instrument_parameters[
    "prt_mode"
]["data"][sweep_slice]

# adjust metadata
radar.metadata = {
    "Conventions": "CF/Radial instrument_parameters",
    "version": "1.2",
    "title": "Py-ART Example PPI CF/Radial file",
    "institution": (
        "United States Department of Energy - Atmospheric "
        "Radiation Measurement (ARM) program"
    ),
    "references": "none",
    "source": "ARM SGP XSAPR Radar",
    "history": "created by jhelmus on evs348532 at 2013-05-22T12:34:56",
    "comment": "none",
    "instrument_name": "xsapr-sgp",
}

pyart.io.write_cfradial("example_cfradial_ppi.nc", radar)



"""
==================================
Create a PPI plot on a cartopy map
==================================

An example which creates a PPI plot of a file with a cartopy background
and range rings

"""

print(__doc__)

# Author: Jason Hemedinger
# License: BSD 3 clause

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import numpy as np

import pyart
from pyart.testing import get_test_data

# Read in the file, create a RadarMapDisplay object
filename = get_test_data("nsaxsaprppiC1.a1.20140201.184802.nc")
radar = pyart.io.read(filename)
display = pyart.graph.RadarMapDisplay(radar)

# Setting projection and ploting the second tilt
projection = ccrs.LambertConformal(
    central_latitude=radar.latitude["data"][0],
    central_longitude=radar.longitude["data"][0],
)

fig = plt.figure(figsize=(6, 6))
display.plot_ppi_map(
    "reflectivity_horizontal",
    1,
    vmin=-20,
    vmax=20,
    min_lon=-157.1,
    max_lon=-156,
    min_lat=71.2,
    max_lat=71.6,
    lon_lines=np.arange(-158, -154, 0.2),
    resolution="10m",
    lat_lines=np.arange(69, 72, 0.1),
    projection=projection,
    fig=fig,
    lat_0=radar.latitude["data"][0],
    lon_0=radar.longitude["data"][0],
)

# Plot range rings at 10, 20, 30, 40km
display.plot_range_ring(10.0, line_style="k-")
display.plot_range_ring(20.0, line_style="k--")
display.plot_range_ring(30.0, line_style="k-")
display.plot_range_ring(40.0, line_style="k--")

# Plot cross hairs
display.plot_line_xy(
    np.array([-40000.0, 40000.0]), np.array([0.0, 0.0]), line_style="k-"
)
display.plot_line_xy(
    np.array([0.0, 0.0]), np.array([-20000.0, 200000.0]), line_style="k-"
)

# Indicate the radar location with a point
display.plot_point(radar.longitude["data"][0], radar.latitude["data"][0])

plt.show()



"""
Reading and RadX Grid objects.

    read_radx_grid

"""

import warnings

import netCDF4

from ..config import get_metadata
from ..core import Grid
from ..io.cfradial import _ncvar_to_dict
from ..io.common import _test_arguments


def read_radx_grid(filename, exclude_fields=None, **kwargs):
    """
    Read a netCDF grid file produced by radx2grid within LROSE.

    Parameters
    ----------
    filename : str
        Filename of netCDF grid file to read.  This file must have been
        produced by :py:func:`write_grid` or have identical layout.

    Other Parameters
    ----------------
    exclude_fields : list
        A list of fields to exclude from the grid object.

    Returns
    -------
    grid : Grid
        Grid object containing gridded data.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    if exclude_fields is None:
        exclude_fields = []

    reserved_variables = [
        "time",
        "x",
        "y",
        "z",
        "origin_latitude",
        "origin_longitude",
        "origin_altitude",
        "point_x",
        "point_y",
        "point_z",
        "projection",
        "point_latitude",
        "point_longitude",
        "point_altitude",
        "radar_latitude",
        "radar_longitude",
        "radar_altitude",
        "radar_name",
        "radar_time",
        "base_time",
        "time_offset",
        "ProjectionCoordinateSystem",
    ]

    dset = netCDF4.Dataset(filename, mode="r")

    # metadata
    metadata = {k: getattr(dset, k) for k in dset.ncattrs()}

    # required reserved variables
    time = _ncvar_to_dict(dset.variables["time"])

    # below is altered from original read_grid to read in origin info from radx file
    # Get dict information from lat array in radx file
    try:
        origin_latitude = _ncvar_to_dict(dset.variables["lat0"])
        # set data single value from Grid info origin from radx file
        origin_latitude["data"] = [
            dset.variables["grid_mapping_0"].latitude_of_projection_origin
        ]
        # Get dict information from lon array in radx file
        origin_longitude = _ncvar_to_dict(dset.variables["lon0"])
        # set data single value from Grid info origin from radx file
        try:
            origin_longitude["data"] = [
                dset.variables["grid_mapping_0"].longitude_of_central_meridian
            ]
        except AttributeError:
            origin_longitude["data"] = [
                dset.variables["grid_mapping_0"].longitude_of_projection_origin
            ]
    except KeyError:
        origin_latitude = get_metadata("latitude")
        origin_latitude["data"] = [
            dset.variables["grid_mapping_0"].latitude_of_projection_origin
        ]

        origin_longitude = get_metadata("longitude")
        try:
            origin_longitude["data"] = [
                dset.variables["grid_mapping_0"].longitude_of_central_meridian
            ]
        except AttributeError:
            origin_longitude["data"] = [
                dset.variables["grid_mapping_0"].longitude_of_projection_origin
            ]

    # only need first alt and it needs to be in meters
    origin_altitude = _ncvar_to_dict(dset.variables["z0"])
    origin_altitude["data"] = [origin_altitude["data"][0] * 1000]
    origin_altitude["units"] = "m"

    # Below is altered from rad_grid to read in the correct variables and
    # change the units of the data in the radx file. Get catersian grid
    # spacing information from radx file and set values and units to meters.
    x = _ncvar_to_dict(dset.variables["x0"])
    x["data"] = x["data"] * 1000
    x["units"] = "m"
    y = _ncvar_to_dict(dset.variables["y0"])
    y["data"] = y["data"] * 1000
    y["units"] = "m"
    z = _ncvar_to_dict(dset.variables["z0"])
    z["data"] = z["data"] * 1000
    z["units"] = "m"

    # Projection
    # Below has been significantly altered form the original rad_grid to
    # obtain the correct info from the radx file to be plotted with Cartopy.
    # Create necessary projection params dict for cartopy and pyart from
    # values in grid_mapping_0 variable in radx created file false easting and
    # northing need to be in meters
    projection = get_grid_projection_dict(dset)

    # map _include_lon_0_lat_0 key to bool type
    if "_include_lon_0_lat_0" in projection:
        v = projection["_include_lon_0_lat_0"]
        projection["_include_lon_0_lat_0"] = {"true": True, "false": False}[v]

    # read in the fields
    fields = {}

    # fields in the file has a shape of (1, nz, ny, nx) with the leading 1
    # indicating time but should shaped (nz, ny, nx) in the Grid object
    field_shape = tuple(len(dset.dimensions[d]) for d in ["z0", "y0", "x0"])
    field_shape_with_time = (1,) + field_shape

    # check all non-reserved variables, those with the correct shape
    # are added to the field dictionary, if a wrong sized field is
    # detected a warning is raised
    field_keys = [k for k in dset.variables if k not in reserved_variables]
    for field in field_keys:
        if field in exclude_fields:
            continue
        field_dic = _ncvar_to_dict(dset.variables[field])
        if field_dic["data"].shape == field_shape_with_time:
            field_dic["data"].shape = field_shape
            fields[field] = field_dic
        else:
            warnings.warn("Field %s skipped due to incorrect shape" % (field))

    # radar_ variables
    if "radar_latitude" in dset.variables:
        radar_latitude = _ncvar_to_dict(dset.variables["radar_latitude"])
    else:
        radar_latitude = None

    if "radar_longitude" in dset.variables:
        radar_longitude = _ncvar_to_dict(dset.variables["radar_longitude"])
    else:
        radar_longitude = None

    if "radar_altitude" in dset.variables:
        radar_altitude = _ncvar_to_dict(dset.variables["radar_altitude"])
    else:
        radar_altitude = None

    if "radar_name" in dset.variables:
        radar_name = _ncvar_to_dict(dset.variables["radar_name"])
    else:
        radar_name = None

    if "radar_time" in dset.variables:
        radar_time = _ncvar_to_dict(dset.variables["radar_time"])
    else:
        radar_time = None

    dset.close()

    return Grid(
        time,
        fields,
        metadata,
        origin_latitude,
        origin_longitude,
        origin_altitude,
        x,
        y,
        z,
        projection=projection,
        radar_latitude=radar_latitude,
        radar_longitude=radar_longitude,
        radar_altitude=radar_altitude,
        radar_name=radar_name,
        radar_time=radar_time,
    )


def get_grid_projection_dict(dset):
    grid_map_name = dset.variables["grid_mapping_0"].grid_mapping_name
    if grid_map_name == "azimuthal_equidistant":
        projstr = "pyart_aeqd"
        projection = {
            "lat_0": dset.variables["grid_mapping_0"].latitude_of_projection_origin,
            "lon_0": dset.variables["grid_mapping_0"].longitude_of_projection_origin,
            "proj": projstr,
            "units": "m",
            "vunits": "m",
            "x_0": dset.variables["grid_mapping_0"].false_easting * 1000,
            "y_0": dset.variables["grid_mapping_0"].false_northing * 1000,
        }

    elif grid_map_name == "transverse_mercator":
        projstr = "tmerc"
        projection = {
            "+k_0": dset.variables["grid_mapping_0"].scale_factor_at_central_meridian,
            "lat_0": dset.variables["grid_mapping_0"].latitude_of_projection_origin,
            "lon_0": dset.variables["grid_mapping_0"].longitude_of_central_meridian,
            "proj": projstr,
            "units": "m",
            "vunits": "m",
            "x_0": dset.variables["grid_mapping_0"].false_easting * 1000,
            "y_0": dset.variables["grid_mapping_0"].false_northing * 1000,
        }

    return projection
